@@ -76,7 +76,6 @@ void endpwent(void);
 #include <string.h>
 #include <stdarg.h>
 #include <malloc.h>
-#include <security/pam_misc.h>
 #include <t_pwd.h>
 
 #ifndef LINUX    /* AGM added this as of 0.2 */
@@ -85,7 +84,7 @@ void endpwent(void);
 #include <security/pam_modules.h>
 #ifdef HAVE_SHADOW_H
 #include <shadow.h>
-#endif
+#endif /* HAVE_SHADOW_H */
 
 #define MAX_PASSWD_TRIES 3
 #define OLD_PASSWORD_PROMPT "Old password: "
