@@ -647,7 +647,7 @@ blkfree(av0)
 
 	while (*av)
 		/* Gross hack to avoid freeing static memory in argbuf - TJW */
-		if(*av > &end)
+		if(*av > (char *)&end)
 			free(*av++);
 		else
 			++av;
