@@ -30,7 +30,7 @@
 
 #include <stdio.h>
 #include "krypto.h"
-#include "krypto_locl.h"
+/*#include "krypto_locl.h"*/
 
 #define NUM_TESTS 34
 static unsigned char ecb_data[NUM_TESTS][8]={
