@@ -39,7 +39,8 @@
  *    of this copyright notice and list of conditions.
  */
 
-#define _POSIX_SOURCE
+/*#define _POSIX_SOURCE*/
+#include <stdio.h>
 #include "t_defines.h"
 
 static int
@@ -92,11 +93,18 @@ t_tohex(dst, src, size)
    int notleading = 0;
 
    register char *chp = dst;
+   *dst = '\0';
    if (size != 0) do {
       if(notleading || *src != '\0') {
+	if(!notleading && (*src & 0xf0) == 0) {
+	  sprintf(chp, "%.1X", * (unsigned char *) src);
+	  chp += 1;
+	}
+	else {
+	  sprintf(chp, "%.2X", * (unsigned char *) src);
+	  chp += 2;
+	}
 	notleading = 1;
-	sprintf(chp, "%.2x", * (unsigned char *) src);
-	chp += 2;
       }
       ++src;
    } while (--size != 0);
