@@ -39,18 +39,41 @@
  *    of this copyright notice and list of conditions.
  */
 
+#include "t_defines.h"
+
+#ifdef HAVE_UNISTD_H
+#include <unistd.h>
+#endif /* HAVE_UNISTD_H */
+
 #include <stdio.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
 
-#include "t_defines.h"
+#ifdef WIN32
+#include <process.h>
+#include <io.h>
+#endif
+
 #include "t_sha.h"
 
 #ifndef NULL
 #define NULL 0
 #endif
 
+#ifdef OPENSSL
+#include <openssl/rand.h>
+static unsigned char ospool[SHA_DIGESTSIZE];
+#elif defined(CRYPTOLIB)
+# include "libcrypt.h"
+static unsigned char crpool[64];
+#else
+static unsigned char randpool[SHA_DIGESTSIZE], randout[SHA_DIGESTSIZE];
+static unsigned long randcnt = 0;
+static unsigned int outpos = 0;
+SHA1_CTX randctxt;
+#endif /* OPENSSL */
+
 /*
  * t_envhash - Generate a 160-bit SHA hash of the environment
  *
@@ -169,13 +192,11 @@ static struct {
   short ppid;
   unsigned char envh[SHA_DIGESTSIZE];
   unsigned char fsh[SHA_DIGESTSIZE];
-  unsigned char devrand[8];
+  unsigned char devrand[20];
   unsigned int trand2;
 } preseed;
 
-static unsigned char randkey1[SHA_DIGESTSIZE], randkey2[SHA_DIGESTSIZE];
-static unsigned char randpool[SHA_DIGESTSIZE];
-static int inpool = 0;
+unsigned long raw_truerand();
 
 static void
 t_initrand()
@@ -186,12 +207,33 @@ t_initrand()
 #else
   struct timeval t;
 #endif
-  int i;
+  int i, r=0;
 
   if(initialized)
     return;
 
-  preseed.trand1 = raw_truerand();
+  initialized = 1;
+
+#if defined(OPENSSL)	/* OpenSSL has nifty win32 entropy-gathering code */
+#if OPENSSL_VERSION_NUMBER >= 0x00905100
+  r = RAND_status();
+#if defined(WINDOWS) || defined(WIN32)
+  if(r)
+    return;
+#else
+#endif
+#endif
+#endif
+
+  i = open("/dev/urandom", O_RDONLY);
+  if(i > 0) {
+    r += read(i, preseed.devrand, sizeof(preseed.devrand));
+    close(i);
+  }
+
+  /* Resort to truerand only if desperate for some Real entropy */
+  if(r == 0)
+    preseed.trand1 = raw_truerand();
 
 #ifdef USE_FTIME
   ftime(&t);
@@ -215,21 +257,27 @@ t_initrand()
   t_envhash(preseed.envh);
   t_fshash(preseed.fsh);
 
-  i = open("/dev/random", O_RDONLY);
-  if(i > 0) {
-    read(i, preseed.devrand, sizeof(preseed.devrand));
-    close(i);
-  }
-  preseed.trand2 = raw_truerand();
+  if(r == 0)
+    preseed.trand2 = raw_truerand();
 
   SHA1Init(&ctxt);
   SHA1Update(&ctxt, (unsigned char *) &preseed, sizeof(preseed));
-  SHA1Final(randkey1, &ctxt);
-  memcpy(randkey2, randkey1, sizeof(randkey1));
-  memset(randpool, 0, sizeof(randpool));
-  inpool = 0;
-
-  initialized = 1;
+#ifdef OPENSSL
+  SHA1Final(ospool, &ctxt);
+  RAND_seed(ospool, sizeof(ospool));
+  memset(ospool, 0, sizeof(ospool));
+#elif defined(CRYPTOLIB)
+  SHA1Final(crpool, &ctxt);
+  for(i = SHA_DIGESTSIZE; i < sizeof(crpool); ++i)
+    crpool[i] = crpool[i - SHA_DIGESTSIZE];
+  seedDesRandom(crpool, sizeof(crpool));
+  memset(crpool, 0, sizeof(crpool));
+#else
+  SHA1Final(randpool, &ctxt);
+  outpos = 0;
+#endif /* OPENSSL */
+  memset((unsigned char *) &preseed, 0, sizeof(preseed));
+  memset((unsigned char *) &ctxt, 0, sizeof(ctxt));
 }
 
 #define NUM_RANDOMS 12
@@ -237,6 +285,9 @@ t_initrand()
 _TYPE( void )
 t_stronginitrand()
 {
+#if 1	/* t_initrand() has been improved enough to make this unnecessary */
+  t_initrand();
+#else
   SHA1_CTX ctxt;
   unsigned int rawrand[NUM_RANDOMS];
   int i;
@@ -249,45 +300,60 @@ t_stronginitrand()
   SHA1Update(&ctxt, (unsigned char *) rawrand, sizeof(rawrand));
   SHA1Final(randkey2, &ctxt);
   memset(rawrand, 0, sizeof(rawrand));
+#endif
 }
 
 /*
- * The strong random number generator.  This uses two 160-bit seeds
- * and uses SHA in a feedback configuration to generate successive
- * outputs.  Let A[0] = 0'160 (160 zero bits).  Then, the outputs
- * become:
- *		A[i+1] = SHA(K1, A[i], K2)
- * where K is the seed and the outputs start with A[1].  Each cycle
- * generates 20 bytes of new output.
+ * The strong random number generator.  This uses a 160-bit seed
+ * and uses SHA-1 in a feedback configuration to generate successive
+ * outputs.  If S[0] is set to the initial seed, then:
+ *
+ *         S[i+1] = SHA-1(i || S[i])
+ *         A[i] = SHA-1(S[i])
+ *
+ * where the A[i] are the output blocks starting with i=0.
+ * Each cycle generates 20 bytes of new output.
  */
 _TYPE( void )
 t_random(data, size)
      unsigned char * data;
      unsigned size;
 {
-  SHA1_CTX ctxt;
-
   if(!initialized)
     t_initrand();
 
-  while(size > inpool) {
-    memcpy(data, randpool + sizeof(randpool) - inpool, inpool);
-    data += inpool;
-    size -= inpool;
+  if(size <= 0)		/* t_random(NULL, 0) forces seed initialization */
+    return;
+
+#ifdef OPENSSL
+  RAND_bytes(data, size);
+#elif defined(CRYPTOLIB)
+  randomBytes(data, size, PSEUDO);
+#else
+  while(size > outpos) {
+    if(outpos > 0) {
+      memcpy(data, randout + (sizeof(randout) - outpos), outpos);
+      data += outpos;
+      size -= outpos;
+    }
 
     /* Recycle */
-    SHA1Init(&ctxt);
-    SHA1Update(&ctxt, randkey1, sizeof(randkey1));
-    SHA1Update(&ctxt, randpool, sizeof(randpool));
-    SHA1Update(&ctxt, randkey2, sizeof(randkey2));
-    SHA1Final(randpool, &ctxt);
-    inpool = sizeof(randpool);
+    SHA1Init(&randctxt);
+    SHA1Update(&randctxt, randpool, sizeof(randpool));
+    SHA1Final(randout, &randctxt);
+    SHA1Init(&randctxt);
+    SHA1Update(&randctxt, (unsigned char *) &randcnt, sizeof(randcnt));
+    SHA1Update(&randctxt, randpool, sizeof(randpool));
+    SHA1Final(randpool, &randctxt);
+    ++randcnt;
+    outpos = sizeof(randout);
   }
 
   if(size > 0) {
-    memcpy(data, randpool + sizeof(randpool) - inpool, size);
-    inpool -= size;
+    memcpy(data, randout + (sizeof(randout) - outpos), size);
+    outpos -= size;
   }
+#endif
 }
 
 /*
