@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1997-1999  The Stanford SRP Authentication Project
+ * Copyright (c) 1997-2000  The Stanford SRP Authentication Project
  * All Rights Reserved.
  *
  * Permission is hereby granted, free of charge, to any person obtaining
@@ -39,14 +39,22 @@
  *    of this copyright notice and list of conditions.
  */
 
+#include "t_defines.h"
+
+#ifdef HAVE_UNISTD_H
+#include <unistd.h>
+#endif /* HAVE_UNISTD_H */
+
 #include <stdio.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #ifdef USE_HOMEDIR
 #include <pwd.h>
 #endif
+#ifdef WIN32
+#include <io.h>
+#endif
 
-#include "t_defines.h"
 #include "t_pwd.h"
 #include "t_read.h"
 #include "t_sha.h"
@@ -118,6 +126,7 @@ t_rewindpw(tpw)
   rewind(tpw->instream);
 }
 
+#ifdef ENABLE_YP
 static void
 savepwent(tpw, pwent)
      struct t_pw * tpw;
@@ -133,6 +142,7 @@ savepwent(tpw, pwent)
   memcpy(tpw->pebuf.salt.data, pwent->salt.data, pwent->salt.len);
   tpw->pebuf.index = pwent->index;
 }
+#endif /* ENABLE_YP */
 
 _TYPE( struct t_pwent * )
 t_getpwent(tpw)
@@ -337,6 +347,8 @@ t_pwcopy(pwdest, pwsrc, diff)
 
   if(diff)
     t_putpwent(diff, pwdest);
+
+  return 0;
 }
 
 static int
@@ -354,6 +366,8 @@ t_pwremove(pwdest, pwsrc, user)
   while((ent = t_getpwent(src)) != NULL)
     if(strcmp(user, ent->name) != 0)
       t_putpwent(ent, pwdest);
+
+  return 0;
 }
 
 /* TODO: Implement a more general method to handle delete/change */
@@ -368,13 +382,17 @@ t_changepw(pwname, diff)
   struct stat st;
   FILE * passfp;
   FILE * bakfp;
+  int rv;
 
   if(pwname == NULL)
     pwname = DEFAULT_PASSWD;
 
-  if((passfp = fopen(pwname, "rb")) == NULL || fstat(fileno(passfp), &st) < 0)
+  if((passfp = fopen(pwname, "rb")) == NULL)
     return -1;
-
+  if(fstat(fileno(passfp), &st) < 0) {
+    fclose(passfp);
+    return -1;
+  }
   if((bakfile = malloc(strlen(pwname) + 5)) == NULL) {
     fclose(passfp);
     return -1;
@@ -392,9 +410,13 @@ t_changepw(pwname, diff)
      (unlink(bakfile2) < 0 || (bakfp = fopen(bakfile2, "wb")) == NULL)) {
     fclose(passfp);
     fclose(bakfp);
+    free(bakfile);
+    free(bakfile2);
     return -1;
   }
 
+  rv = 0;
+
   if(syspw != NULL) {
     t_closepw(syspw);
     syspw = NULL;
@@ -414,9 +436,11 @@ t_changepw(pwname, diff)
 #ifdef USE_RENAME
   unlink(bakfile);
   if(rename(pwname, bakfile) < 0)
-    return -1;
-  if(rename(bakfile2, pwname) < 0)
-    return -1;
+    rv = -1;
+  else {
+    if(rename(bakfile2, pwname) < 0)
+      rv = -1;
+  }
 #else
   unlink(bakfile);
   link(pwname, bakfile);
@@ -427,7 +451,7 @@ t_changepw(pwname, diff)
   free(bakfile);
   free(bakfile2);
 
-  return 0;
+  return rv;
 }
 
 _TYPE( int )
