@@ -203,6 +203,8 @@ t_servergetkey(ts, clientval)
   SHA1Update(&ctxt, ts->B.data, ts->B.len);
   SHA1Final(dig, &ctxt);
   u = (dig[0] << 24) | (dig[1] << 16) | (dig[2] << 8) | dig[3];
+  if(u == 0)		/* Check for u == 0 */
+    return NULL;
 
   SHA1Update(&ts->oldhash, clientval->data, clientval->len);
   SHA1Update(&ts->hash, clientval->data, clientval->len);
@@ -221,7 +223,7 @@ t_servergetkey(ts, clientval)
   BigIntegerFree(v);
   BigIntegerFree(prod);
 
-  if(BigIntegerCmpInt(res, 1) <= 0) {	/* Check for Av^u == 1 (mod n) */
+  if(BigIntegerCmpInt(res, 1) <= 0) {	/* Check for Av^u == 0,1 (mod n) */
     BigIntegerFree(res);
     BigIntegerFree(b);
     BigIntegerFree(n);
