@@ -21,7 +21,7 @@
 
 #undef AUTHENTICATION
 
-#undef SRP
+#undef HAVE_SRP
 
 #undef KRB4
 
