@@ -62,6 +62,7 @@
 
 #ifdef	ENCRYPTION
 void encrypt_init P((char *, int));
+void encrypt_reset P((void));
 Encryptions *findencryption P((int));
 void encrypt_send_supprt P((void));
 void encrypt_auto P((int));
@@ -95,7 +96,7 @@ int krbdes_is P((unsigned char *, int));
 int krbdes_reply P((unsigned char *, int));
 void krbdes_init P((int));
 int krbdes_start P((int, int));
-void krbdes_session P((Session_Key *, int));
+int krbdes_session P((Session_Key *, int));
 void krbdes_printsub P((unsigned char *, int, unsigned char *, int));
 
 void cfb64_encrypt P((unsigned char *, int));
@@ -104,7 +105,7 @@ void cfb64_init P((int));
 int cfb64_start P((int, int));
 int cfb64_is P((unsigned char *, int));
 int cfb64_reply P((unsigned char *, int));
-void cfb64_session P((Session_Key *, int));
+int cfb64_session P((Session_Key *, int));
 int cfb64_keyid P((int, unsigned char *, int *));
 void cfb64_printsub P((unsigned char *, int, unsigned char *, int));
 
@@ -114,7 +115,7 @@ void ofb64_init P((int));
 int ofb64_start P((int, int));
 int ofb64_is P((unsigned char *, int));
 int ofb64_reply P((unsigned char *, int));
-void ofb64_session P((Session_Key *, int));
+int ofb64_session P((Session_Key *, int));
 int ofb64_keyid P((int, unsigned char *, int *));
 void ofb64_printsub P((unsigned char *, int, unsigned char *, int));
 
@@ -124,7 +125,7 @@ void des3_cfb64_init P((int));
 int  des3_cfb64_start P((int, int));
 int  des3_cfb64_is P((unsigned char *, int));
 int  des3_cfb64_reply P((unsigned char *, int));
-void des3_cfb64_session P((Session_Key *, int));
+int  des3_cfb64_session P((Session_Key *, int));
 int  des3_cfb64_keyid P((int, unsigned char *, int *));
 void des3_cfb64_printsub P((unsigned char *, int, unsigned char *, int));
 
@@ -134,7 +135,7 @@ void des3_ofb64_init P((int));
 int  des3_ofb64_start P((int, int));
 int  des3_ofb64_is P((unsigned char *, int));
 int  des3_ofb64_reply P((unsigned char *, int));
-void des3_ofb64_session P((Session_Key *, int));
+int  des3_ofb64_session P((Session_Key *, int));
 int  des3_ofb64_keyid P((int, unsigned char *, int *));
 void des3_ofb64_printsub P((unsigned char *, int, unsigned char *, int));
 
@@ -144,7 +145,7 @@ void cast_cfb64_init P((int));
 int cast_cfb64_start P((int, int));
 int cast_cfb64_is P((unsigned char *, int));
 int cast_cfb64_reply P((unsigned char *, int));
-void cast_cfb64_session P((Session_Key *, int));
+int cast_cfb64_session P((Session_Key *, int));
 int cast_cfb64_keyid P((int, unsigned char *, int *));
 void cast_cfb64_printsub P((unsigned char *, int, unsigned char *, int));
 
@@ -154,7 +155,7 @@ void cast_ofb64_init P((int));
 int cast_ofb64_start P((int, int));
 int cast_ofb64_is P((unsigned char *, int));
 int cast_ofb64_reply P((unsigned char *, int));
-void cast_ofb64_session P((Session_Key *, int));
+int cast_ofb64_session P((Session_Key *, int));
 int cast_ofb64_keyid P((int, unsigned char *, int *));
 void cast_ofb64_printsub P((unsigned char *, int, unsigned char *, int));
 
@@ -164,7 +165,7 @@ void castexp_cfb64_init P((int));
 int castexp_cfb64_start P((int, int));
 int castexp_cfb64_is P((unsigned char *, int));
 int castexp_cfb64_reply P((unsigned char *, int));
-void castexp_cfb64_session P((Session_Key *, int));
+int castexp_cfb64_session P((Session_Key *, int));
 int castexp_cfb64_keyid P((int, unsigned char *, int *));
 void castexp_cfb64_printsub P((unsigned char *, int, unsigned char *, int));
 
@@ -174,20 +175,28 @@ void castexp_ofb64_init P((int));
 int castexp_ofb64_start P((int, int));
 int castexp_ofb64_is P((unsigned char *, int));
 int castexp_ofb64_reply P((unsigned char *, int));
-void castexp_ofb64_session P((Session_Key *, int));
+int castexp_ofb64_session P((Session_Key *, int));
 int castexp_ofb64_keyid P((int, unsigned char *, int *));
 void castexp_ofb64_printsub P((unsigned char *, int, unsigned char *, int));
 
-#ifdef NEWDESLIB
+#ifdef DESLIB3
 #define des_fixup_key_parity des_set_odd_parity
-#define des_new_random_key des_random_key
-#define des_set_random_generator_seed des_random_seed
-#endif
+#endif /* DESLIB3 */
 
+#ifndef DESLIB4
+#define des_new_random_key des_random_key
+#define des_set_random_generator_seed(x)
+#define des_init_random_number_generator(x)
+#endif /* !DESLIB4 */
+
+/* Try to avoid conflict with existing DES headers */
+#if !defined(DES_DEFS) && !defined(HEADER_DES_H)
 int  des_new_random_key P((Block));
 void des_set_random_generator_seed P((Block));
 int des_key_sched P((Block, Schedule));
 void des_ecb_encrypt P((Block, Block, Schedule, int));
 int  des_string_to_key P((char *, Block));
 void des_fixup_key_parity P((Block));
+#endif /* !DES_DEFS && !HEADER_DES_H */
+
 #endif	/* ENCRYPTION */
