@@ -35,6 +35,10 @@
 static char sccsid[] = "@(#)enc_des.c	8.3 (Berkeley) 5/30/95";
 #endif /* not lint */
 
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
+
 #ifdef	ENCRYPTION
 # ifdef	AUTHENTICATION
 #  ifdef DES_ENCRYPTION
@@ -44,6 +48,13 @@ static char sccsid[] = "@(#)enc_des.c	8.3 (Berkeley) 5/30/95";
 #include <stdlib.h>
 #endif
 
+#ifdef OPENSSL_DES
+#include <openssl/rand.h>
+#include <openssl/des.h>
+#endif
+#ifdef CRYPTOLIB_DES
+#include "libcrypt.h"
+#endif
 #include "encrypt.h"
 #include "key-proto.h"
 #include "misc-proto.h"
@@ -115,7 +126,7 @@ void fb64_init P((struct fb *));
 static int fb64_start P((struct fb *, int, int));
 int fb64_is P((unsigned char *, int, struct fb *));
 int fb64_reply P((unsigned char *, int, struct fb *));
-static void fb64_session P((Session_Key *, int, struct fb *));
+static int fb64_session P((Session_Key *, int, struct fb *));
 void fb64_stream_key P((Block, struct stinfo *));
 int fb64_keyid P((int, unsigned char *, int *, struct fb *));
 
@@ -215,9 +226,18 @@ fb64_start(fbp, dir, server)
 		/*
 		 * Create a random feed and send it over.
 		 */
+#ifdef CRYPTOLIB_DES
+		randomBytes(fbp->temp_feed, sizeof(Block), PSEUDO);
+#else
 		des_new_random_key(fbp->temp_feed);
+#endif
+#ifdef CRYPTOLIB_DES
+		block_cipher((unsigned char *)fbp->krbdes_sched,
+			     fbp->temp_feed, 0);
+#else
 		des_ecb_encrypt(fbp->temp_feed, fbp->temp_feed,
 				fbp->krbdes_sched, 1);
+#endif
 		p = fbp->fb_feed + 3;
 		*p++ = ENCRYPT_IS;
 		p++;
@@ -266,7 +286,6 @@ fb64_is(data, cnt, fbp)
 {
 	int x;
 	unsigned char *p;
-	Block b;
 	register int state = fbp->state[DIR_DECRYPT-1];
 
 	if (cnt-- < 1)
@@ -358,7 +377,6 @@ fb64_reply(data, cnt, fbp)
 {
 	int x;
 	unsigned char *p;
-	Block b;
 	register int state = fbp->state[DIR_ENCRYPT-1];
 
 	if (cnt-- < 1)
@@ -393,23 +411,23 @@ fb64_reply(data, cnt, fbp)
 	return(fbp->state[DIR_ENCRYPT-1] = state);
 }
 
-	void
+	int
 cfb64_session(key, server)
 	Session_Key *key;
 	int server;
 {
-	fb64_session(key, server, &fb[CFB]);
+	return fb64_session(key, server, &fb[CFB]);
 }
 
-	void
+	int
 ofb64_session(key, server)
 	Session_Key *key;
 	int server;
 {
-	fb64_session(key, server, &fb[OFB]);
+	return fb64_session(key, server, &fb[OFB]);
 }
 
-	static void
+	static int
 fb64_session(key, server, fbp)
 	Session_Key *key;
 	int server;
@@ -418,6 +436,9 @@ fb64_session(key, server, fbp)
 	int usehigh;
 	struct stinfo * hstream;
 	struct stinfo * lstream;
+#ifdef CRYPTOLIB_DES
+	unsigned char des_seed[64];
+#endif
 
 	if(server) {
 	  hstream = &fbp->streams[DIR_ENCRYPT-1];
@@ -432,30 +453,45 @@ fb64_session(key, server, fbp)
 		if (encrypt_debug_mode)
 			printf("Can't set krbdes's session key (%d < %d)\r\n",
 				key ? key->length : 0, sizeof(Block));
-		return;
+		return -1;
 	}
 	usehigh = (key->type == SK_DES || key->length < 2 * sizeof(Block)) ?
 		    0 : 1;
 	if(usehigh) {
 	  memmove((void *) fbp->krbdes_key,
 		  (void *) (key->data + sizeof(Block)), sizeof(Block));
+#ifndef CRYPTOLIB_DES
 	  des_fixup_key_parity(fbp->krbdes_key);
+#endif
 	  fb64_stream_key(fbp->krbdes_key, hstream);
 	}
 
 	memmove((void *)fbp->krbdes_key, (void *)key->data, sizeof(Block));
+#ifndef CRYPTOLIB_DES
 	if (key->type != SK_DES)
 	  des_fixup_key_parity(fbp->krbdes_key);
+#endif
 
 	if(!usehigh)
 	  fb64_stream_key(fbp->krbdes_key, hstream);
 	fb64_stream_key(fbp->krbdes_key, lstream);
 
 	if (fbp->once == 0) {
-		des_set_random_generator_seed(fbp->krbdes_key);
+#ifdef OPENSSL_DES
+		des_random_seed(fbp->krbdes_key);
+#elif defined(CRYPTOLIB_DES)
+		memmove((void *)des_seed, fbp->krbdes_key, sizeof(Block));
+		seedDesRandom(des_seed, sizeof(des_seed));
+#elif defined(DESLIB4)
+		des_init_random_number_generator(fbp->krbdes_key);
+#endif
 		fbp->once = 1;
 	}
+#ifdef CRYPTOLIB_DES
+	key_setup(fbp->krbdes_key, (unsigned char *)fbp->krbdes_sched);
+#else
 	des_key_sched(fbp->krbdes_key, fbp->krbdes_sched);
+#endif
 	/*
 	 * Now look to see if krbdes_start() was was waiting for
 	 * the key to show up.  If so, go ahead an call it now
@@ -465,6 +501,7 @@ fb64_session(key, server, fbp)
 		fbp->need_start = 0;
 		fb64_start(fbp, DIR_ENCRYPT, server);
 	}
+        return 0;
 }
 
 /*
@@ -575,7 +612,11 @@ fb64_stream_iv(seed, stp)
 	memmove((void *)stp->str_iv, (void *)seed, sizeof(Block));
 	memmove((void *)stp->str_output, (void *)seed, sizeof(Block));
 
+#ifdef CRYPTOLIB_DES
+	key_setup(stp->str_ikey, (unsigned char *)stp->str_sched);
+#else
 	des_key_sched(stp->str_ikey, stp->str_sched);
+#endif
 
 	stp->str_index = sizeof(Block);
 }
@@ -586,7 +627,11 @@ fb64_stream_key(key, stp)
 	register struct stinfo *stp;
 {
 	memmove((void *)stp->str_ikey, (void *)key, sizeof(Block));
+#ifdef CRYPTOLIB_DES
+	key_setup(key, (unsigned char *)stp->str_sched);
+#else
 	des_key_sched(key, stp->str_sched);
+#endif
 
 	memmove((void *)stp->str_output, (void *)stp->str_iv, sizeof(Block));
 
@@ -626,9 +671,14 @@ cfb64_encrypt(s, c)
 	index = stp->str_index;
 	while (c-- > 0) {
 		if (index == sizeof(Block)) {
-			Block b;
-			des_ecb_encrypt(stp->str_output, b, stp->str_sched, 1);
-			memmove((void *)stp->str_feed, (void *)b, sizeof(Block));
+#ifdef CRYPTOLIB_DES
+			memmove((void *)stp->str_feed, (void *)stp->str_output,
+				sizeof(Block));
+			block_cipher((unsigned char *)stp->str_sched,
+				     stp->str_feed, 0);
+#else
+			des_ecb_encrypt(stp->str_output, stp->str_feed, stp->str_sched, 1);
+#endif
 			index = 0;
 		}
 
@@ -660,9 +710,14 @@ cfb64_decrypt(data)
 
 	index = stp->str_index++;
 	if (index == sizeof(Block)) {
-		Block b;
-		des_ecb_encrypt(stp->str_output, b, stp->str_sched, 1);
-		memmove((void *)stp->str_feed, (void *)b, sizeof(Block));
+#ifdef CRYPTOLIB_DES
+		memmove((void *)stp->str_feed, (void *)stp->str_output,
+			sizeof(Block));
+		block_cipher((unsigned char *)stp->str_sched,
+			     stp->str_feed, 0);
+#else
+		des_ecb_encrypt(stp->str_output, stp->str_feed, stp->str_sched, 1);
+#endif
 		stp->str_index = 1;	/* Next time will be 1 */
 		index = 0;		/* But now use 0 */
 	}
@@ -702,9 +757,12 @@ ofb64_encrypt(s, c)
 	index = stp->str_index;
 	while (c-- > 0) {
 		if (index == sizeof(Block)) {
-			Block b;
-			des_ecb_encrypt(stp->str_feed, b, stp->str_sched, 1);
-			memmove((void *)stp->str_feed, (void *)b, sizeof(Block));
+#ifdef CRYPTOLIB_DES
+			block_cipher((unsigned char *)stp->str_sched,
+				     stp->str_feed, 0);
+#else
+			des_ecb_encrypt(stp->str_feed, stp->str_feed, stp->str_sched, 1);
+#endif
 			index = 0;
 		}
 		*s++ ^= stp->str_feed[index];
@@ -733,9 +791,12 @@ ofb64_decrypt(data)
 
 	index = stp->str_index++;
 	if (index == sizeof(Block)) {
-		Block b;
-		des_ecb_encrypt(stp->str_feed, b, stp->str_sched, 1);
-		memmove((void *)stp->str_feed, (void *)b, sizeof(Block));
+#ifdef CRYPTOLIB_DES
+		block_cipher((unsigned char *)stp->str_sched,
+			     stp->str_feed, 0);
+#else
+		des_ecb_encrypt(stp->str_feed, stp->str_feed, stp->str_sched, 1);
+#endif
 		stp->str_index = 1;	/* Next time will be 1 */
 		index = 0;		/* But now use 0 */
 	}
