@@ -30,6 +30,11 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  */
+/*
+ * START_TLS and FORWARD_X support from Peter Runestig <peter@runestig.com>
+ * and Jeffrey Altman <jaltman@columbia.edu> incorporated into the
+ * SRP Telnet distribution 10/19/2000 by Tom Wu <tjw@cs.stanford.edu>.
+ */
 
 #ifndef lint
 static char sccsid[] = "@(#)sys_bsd.c	8.4 (Berkeley) 5/30/95";
@@ -40,10 +45,29 @@ static char sccsid[] = "@(#)sys_bsd.c	8.4 (Berkeley) 5/30/95";
  * (at least between 4.x and dos) which is used in telnet.c.
  */
 
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
+
+#include <unistd.h>
+
+#ifdef TLS
+#define RECV tls_recv
+#define READ tls_read
+#include "tlsutil.h"
+#else
+#define RECV recv
+#define READ read
+#endif
 
-#include <fcntl.h>
 #include <sys/types.h>
 #include <sys/time.h>
+
+#ifdef FWD_X
+#include "fwdxutil.h"
+#endif
+
+#include <fcntl.h>
 #include <sys/socket.h>
 #include <signal.h>
 #include <errno.h>
@@ -140,7 +164,7 @@ TerminalWrite(buf, n)
 
     int
 TerminalRead(buf, n)
-    unsigned char *buf;
+    char *buf;
     int  n;
 {
     return read(tin, buf, n);
@@ -669,7 +693,7 @@ TerminalNewMode(f)
 #endif
 #ifdef	SIGTSTP
 	(void) signal(SIGTSTP, SIG_DFL);
-# ifndef SOLARIS
+# if !defined(SOLARIS) && !defined(SCO)
 	(void) sigsetmask(sigblock(0) & ~(1<<(SIGTSTP-1)));
 # else	SOLARIS
 	(void) sigrelse(SIGTSTP);
@@ -1002,6 +1026,12 @@ process_rings(netin, netout, netex, ttyin, ttyout, poll)
     int returnValue = 0;
     static struct timeval TimeValue = { 0 };
 
+    FD_ZERO(&ibits);
+#ifdef COMMENT
+    FD_ZERO(&obits);
+    FD_ZERO(&xbits);
+#endif
+
     if (netout) {
 	FD_SET(net, &obits);
     }
@@ -1029,7 +1059,13 @@ process_rings(netin, netout, netex, ttyin, ttyout, poll)
     if (netex) {
 	FD_SET(net, &xbits);
     }
+
+#ifdef FWD_X
+    fwdx_init_fd_set(&ibits);
+    if ((c = select(fwdx_max_socket() + 16, &ibits, &obits, &xbits,
+#else
     if ((c = select(16, &ibits, &obits, &xbits,
+#endif
 			(poll == 0)? (struct timeval *)0 : &TimeValue)) < 0) {
 	if (c == -1) {
 		    /*
@@ -1122,9 +1158,9 @@ process_rings(netin, netout, netex, ttyin, ttyout, poll)
 
 	    ioctl(net, SIOCATMARK, (char *)&atmark);
 	    if (atmark) {
-		c = recv(net, netiring.supply, canread, MSG_OOB);
+		c = RECV(net, netiring.supply, canread, MSG_OOB);
 		if ((c == -1) && (errno == EINVAL)) {
-		    c = recv(net, netiring.supply, canread, 0);
+		    c = RECV(net, netiring.supply, canread, 0);
 		    if (clocks.didnetreceive < clocks.gotDM) {
 			SYNCHing = stilloob(net);
 		    }
@@ -1141,7 +1177,7 @@ process_rings(netin, netout, netex, ttyin, ttyout, poll)
 		     * increment the count...)
 		     */
 		    int i;
-		    i = recv(net, netiring.supply + c, canread - c, MSG_OOB);
+		    i = RECV(net, netiring.supply + c, canread - c, MSG_OOB);
 		    if (i == c &&
 			 memcmp(netiring.supply, netiring.supply + c, i) == 0) {
 			bogus_oob = 1;
@@ -1159,21 +1195,21 @@ process_rings(netin, netout, netex, ttyin, ttyout, poll)
 		     * to clear the atmark to get out of
 		     * an infinate loop.
 		     */
-		    i = read(net, netiring.supply + c, canread - c);
+		    i = READ(net, netiring.supply + c, canread - c);
 		    if (i > 0)
 			c += i;
 		}
 	    } else {
-		c = recv(net, netiring.supply, canread, 0);
+		c = RECV(net, netiring.supply, canread, 0);
 	    }
 	} else {
-	    c = recv(net, netiring.supply, canread, 0);
+	    c = RECV(net, netiring.supply, canread, 0);
 	}
 	settimer(didnetreceive);
 #else	/* !defined(SO_OOBINLINE) */
-	c = recv(net, (char *)netiring.supply, canread, 0);
+	c = RECV(net, (char *)netiring.supply, canread, 0);
 #endif	/* !defined(SO_OOBINLINE) */
-	if (c < 0 && errno == EWOULDBLOCK) {
+	if (c < 0 && (errno == EWOULDBLOCK || errno == EINTR)) {
 	    c = 0;
 	} else if (c <= 0) {
 	    return -1;
@@ -1186,6 +1222,14 @@ process_rings(netin, netout, netex, ttyin, ttyout, poll)
 	returnValue = 1;
     }
 
+#ifdef FWD_X
+    /*
+     * Any "X-server" connections / data waiting?
+     */
+    if (fwdx_check_sockets(&ibits))
+	returnValue = 1;
+#endif
+
     /*
      * Something to read from the tty...
      */
