@@ -30,12 +30,22 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  */
+/*
+ * START_TLS and FORWARD_X support from Peter Runestig <peter@runestig.com>
+ * and Jeffrey Altman <jaltman@columbia.edu> incorporated into the
+ * SRP Telnet distribution 10/19/2000 by Tom Wu <tjw@cs.stanford.edu>.
+ */
 
 #ifndef lint
 static char sccsid[] = "@(#)telnet.c	8.4 (Berkeley) 5/30/95";
 #endif /* not lint */
 
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
+
 #include <sys/types.h>
+#include <sys/time.h>
 
 #if	defined(unix)
 #include <signal.h>
@@ -56,9 +66,33 @@ static char sccsid[] = "@(#)telnet.c	8.4 (Berkeley) 5/30/95";
 #include "types.h"
 #include "general.h"
 
+#if	!defined(__P)
+#ifdef	__STDC__
+#define __P(x)	x
+#else
+#define __P(x)	()
+#endif
+#endif
+
+#ifdef TLS
+#define RECV tls_recv
+#include <errno.h>
+#include <netinet/in.h>
+#include "tlsutil.h"
+void netringflush __P((void));
+void announce_my_options __P((void));
+#else
+#define RECV recv
+#endif
+#ifdef FWD_X
+#include "fwdxutil.h"
+#endif
 
 #define	strip(x) ((my_want_state_is_wont(TELOPT_BINARY)) ? ((x)&0x7f) : (x))
 
+void ExitString __P((char *string, int returnCode));
+int process_rings __P((int netin, int netout, int netex, int ttyin, int ttyout, int poll));
+
 static unsigned char	subbuffer[SUBBUFSIZE],
 			*subpointer, *subend;	 /* buffer for sub-options */
 #define	SB_CLEAR()	subpointer = subbuffer;
@@ -93,7 +127,7 @@ int
 	noasynchnet = 0,/* User specified "-noasynch" on command line */
 	askedSGA = 0,	/* We have talked about suppress go ahead */
 #endif	/* defined(TN3270) */
-	telnetport,
+	telnetport = 0,
         wantencryption = 0,
 	SYNCHing,	/* we are in TELNET SYNCH mode */
 	flushout,	/* flush output */
@@ -108,6 +142,7 @@ int
 	globalmode;
 
 char *prompt = 0;
+int scheduler_lockout_tty = 0;
 
 cc_t escape;
 cc_t rlogin;
@@ -128,7 +163,7 @@ cc_t echoc;
 #define	TS_SB		7		/* sub-option collection */
 #define	TS_SE		8		/* looking for sub-option end */
 
-static int	telrcv_state;
+static int	telrcv_state = TS_DATA;
 #ifdef	OLD_ENVIRON
 unsigned char telopt_environ = TELOPT_NEW_ENVIRON;
 #else
@@ -169,15 +204,13 @@ Modelist modelist[] = {
  * Initialize telnet environment.
  */
 
-    void
-init_telnet()
+void
+reset_telnet()
 {
-    env_init();
-
     SB_CLEAR();
     ClearArray(options);
 
-    connected = In3270 = ISend = localflow = donebinarytoggle = 0;
+    In3270 = ISend = localflow = donebinarytoggle = 0;
 #if	defined(AUTHENTICATION) || defined(ENCRYPTION)
     auth_encrypt_connect(connected);
 #endif	/* defined(AUTHENTICATION) || defined(ENCRYPTION)  */
@@ -198,11 +231,19 @@ init_telnet()
 }
 
 
+void
+init_telnet()
+{
+    connected = 0;
+    env_init();
+    reset_telnet();
+}
+
 #ifdef	notdef
 #include <varargs.h>
 
     /*VARARGS*/
-    static void
+static void
 printring(va_alist)
     va_dcl
 {
@@ -256,7 +297,7 @@ printring(va_alist)
  * is in disagreement as to what the current state should be.
  */
 
-    void
+void
 send_do(c, init)
     register int c, init;
 {
@@ -272,7 +313,7 @@ send_do(c, init)
     printoption("SENT", DO, c);
 }
 
-    void
+void
 send_dont(c, init)
     register int c, init;
 {
@@ -288,7 +329,7 @@ send_dont(c, init)
     printoption("SENT", DONT, c);
 }
 
-    void
+void
 send_will(c, init)
     register int c, init;
 {
@@ -304,7 +345,7 @@ send_will(c, init)
     printoption("SENT", WILL, c);
 }
 
-    void
+void
 send_wont(c, init)
     register int c, init;
 {
@@ -321,7 +362,7 @@ send_wont(c, init)
 }
 
 
-	void
+void
 willoption(option)
 	int option;
 {
@@ -370,11 +411,23 @@ willoption(option)
 #if	defined(AUTHENTICATION)
 	    case TELOPT_AUTHENTICATION:
 #endif
-#ifdef	ENCRYPTION
-	    case TELOPT_ENCRYPT:
-#endif /* ENCRYPTION */
 		new_state_ok = 1;
 		break;
+#ifdef	ENCRYPTION
+	    case TELOPT_ENCRYPT:
+#  ifdef TLS
+		if(tls_active)		/* Refuse ENCRYPT if under TLS */
+			break;
+#  endif
+		new_state_ok = 1;
+		break;
+#endif /* ENCRYPTION */
+
+#ifdef	FWD_X
+	    case TELOPT_FORWARD_X:	/* TODO: refuse if not ENCRYPT|TLS */
+		new_state_ok = 1;
+		break;
+#endif /* FWD_X */
 
 	    case TELOPT_TM:
 		if (flushout)
@@ -403,12 +456,38 @@ willoption(option)
 	}
 	set_my_state_do(option);
 #ifdef	ENCRYPTION
-	if (option == TELOPT_ENCRYPT)
+	if (option == TELOPT_ENCRYPT
+#  ifdef TLS
+	    && !tls_active
+#  endif
+	    )
 		encrypt_send_support();
 #endif	/* ENCRYPTION */
+#ifdef  FWD_X
+	if (option == TELOPT_FORWARD_X && fwdx_enable_flag) {
+	   int screen = fwdx_get_screen_no(getenv("DISPLAY"));
+	    if (!fwdx_sbdata)
+		fwdx_sbdata = malloc(fwdx_sbdata_size);
+	    if (fwdx_sbdata && getenv("DISPLAY") && screen >= 0) {
+		unsigned char sb[] = { IAC, SB, TELOPT_FORWARD_X, FWDX_SCREEN, 0, IAC, SE };
+		sb[4] = (unsigned char)screen;
+		set_my_want_state_do(option);
+		send_do(option, 1);
+		if (sizeof(sb) < NETROOM()) {
+		    ring_supply_data(&netoring, sb, sizeof(sb));
+		    printsub('>', sb + 2, sizeof(sb) - 2);
+		}
+/*@*/	        else printf("telopt_fwdx_screen: not enough room in buffer\n");
+	    }
+	    else {
+	 	do_dont_resp[option]++;
+		send_dont(option, 0);
+	    }
+	}
+#endif /* FWD_X */
 }
 
-	void
+void
 wontoption(option)
 	int option;
 {
@@ -457,7 +536,7 @@ wontoption(option)
 	set_my_state_dont(option);
 }
 
-	static void
+static void
 dooption(option)
 	int option;
 {
@@ -484,6 +563,26 @@ dooption(option)
 		set_my_state_wont(TELOPT_TM);
 		return;
 
+#ifdef TLS
+	    case TELOPT_START_TLS: {
+		char buf[] = { IAC, WILL, TELOPT_START_TLS, IAC, SB, TELOPT_START_TLS,
+				TLS_FOLLOWS, IAC, SE };
+		if(!tls_enabled)
+		    break;
+		if (tls_init())
+		    ExitString("Error initalizing OpenSSL.\n", 1);
+		/* usage: process_rings(netin, netout, netex, ttyin, ttyout, !block) */
+		process_rings(0, 1, 1, 0, 0, 1);
+		tls_send(net, buf, sizeof(buf), 0);
+		printoption("SENT", WILL, TELOPT_START_TLS);
+		printsub('>', &buf[5], 4);
+		tls_suspend_iacs = 1;
+		set_my_want_state_will(option);
+		set_my_state_will(option);
+		return;
+		}
+#endif
+
 #	if defined(TN3270)
 	    case TELOPT_EOR:		/* end of record */
 #	endif	/* defined(TN3270) */
@@ -493,11 +592,17 @@ dooption(option)
 	    case TELOPT_LFLOW:		/* local flow control */
 	    case TELOPT_TTYPE:		/* terminal type option */
 	    case TELOPT_SGA:		/* no big deal */
-#ifdef	ENCRYPTION
-	    case TELOPT_ENCRYPT:	/* encryption variable option */
-#endif	/* ENCRYPTION */
 		new_state_ok = 1;
 		break;
+#ifdef	ENCRYPTION
+	    case TELOPT_ENCRYPT:	/* encryption variable option */
+#  ifdef TLS
+		if(tls_active)		/* Refuse ENCRYPT if under TLS */
+			break;
+#  endif
+		new_state_ok = 1;
+		break;
+#endif	/* ENCRYPTION */
 
 	    case TELOPT_NEW_ENVIRON:	/* New environment variable option */
 #ifdef	OLD_ENVIRON
@@ -556,6 +661,24 @@ dooption(option)
 	     * other side has acknowledged the option.
 	     */
 	    switch (option) {
+#ifdef TLS
+	    case TELOPT_START_TLS: {
+		char buf[] = { IAC, SB, TELOPT_START_TLS,
+				TLS_FOLLOWS, IAC, SE };
+		if(!tls_enabled)
+		    break;
+		if (tls_init())
+		    ExitString("Error initalizing OpenSSL.\n", 1);
+		/* usage: process_rings(netin, netout, netex, ttyin, ttyout, !block) */
+		process_rings(0, 1, 1, 0, 0, 1);
+		tls_send(net, buf, sizeof(buf), 0);
+		printsub('>', &buf[2], 4);
+		tls_suspend_iacs = 1;
+		set_my_state_will(option);
+		return;
+		}
+#endif
+
 	    case TELOPT_LINEMODE:
 #ifdef	KLUDGELINEMODE
 		kludgelinemode = 0;
@@ -571,7 +694,7 @@ dooption(option)
 	set_my_state_will(option);
 }
 
-	static void
+static void
 dontoption(option)
 	int option;
 {
@@ -596,6 +719,17 @@ dontoption(option)
 		send_will(TELOPT_OLD_ENVIRON, 1);
 		telopt_environ = TELOPT_OLD_ENVIRON;
 		break;
+#endif
+#ifdef	AUTHENTICATION
+	    case TELOPT_AUTHENTICATION:
+#ifdef  TLS	/* TJW: experimental */
+		if(my_state_is_will(TELOPT_START_TLS) &&
+		   !my_want_state_is_wont(TELOPT_START_TLS) &&
+		   !tls_active && tls_enabled)
+		    break;
+#endif
+		auth_warn();
+		break;
 #endif
 	    }
 	    /* we always accept a DONT */
@@ -617,7 +751,7 @@ dontoption(option)
 static char *name_unknown = "UNKNOWN";
 static char *unknown[] = { 0, 0 };
 
-	char **
+char **
 mklist(buf, name)
 	char *buf, *name;
 {
@@ -728,7 +862,7 @@ mklist(buf, name)
 		return(unknown);
 }
 
-	int
+int
 is_unique(name, as, ae)
 	register char *name, **as, **ae;
 {
@@ -743,16 +877,16 @@ is_unique(name, as, ae)
 }
 
 #ifdef	TERMCAP
-char termbuf[1024];
+char termbuf[4096];
 
 	/*ARGSUSED*/
-	int
+int
 t_setupterm(tname, fd, errp)
 	char *tname;
 	int fd, *errp;
 {
 	if (tgetent(termbuf, tname) == 1) {
-		termbuf[1023] = '\0';
+		termbuf[sizeof(termbuf) - 1] = '\0';
 		if (errp)
 			*errp = 1;
 		return(0);
@@ -764,11 +898,20 @@ t_setupterm(tname, fd, errp)
 #else
 #define	termbuf	ttytype
 extern char ttytype[];
+
+	/*ARGSUSED*/
+	int
+t_setupterm(tname, fd, errp)
+	char *tname;
+	int fd, *errp;
+{
+  return setupterm(tname, fd, errp);
+}
 #endif
 
 int resettermname = 1;
 
-	char *
+char *
 gettermname()
 {
 	char *tname;
@@ -811,7 +954,7 @@ gettermname()
  *		Linemode
  */
 
-    static void
+static void
 suboption()
 {
     unsigned char subchar;
@@ -836,8 +979,9 @@ suboption()
 	    name = gettermname();
 	    len = strlen(name) + 4 + 2;
 	    if (len < NETROOM()) {
-		sprintf((char *)temp, "%c%c%c%c%s%c%c", IAC, SB, TELOPT_TTYPE,
-				TELQUAL_IS, name, IAC, SE);
+		snprintf((char *)temp, sizeof(temp),
+			 "%c%c%c%c%s%c%c", IAC, SB, TELOPT_TTYPE,
+			 TELQUAL_IS, name, IAC, SE);
 		ring_supply_data(&netoring, temp, len);
 		printsub('>', &temp[2], len-2);
 	    } else {
@@ -846,6 +990,93 @@ suboption()
 	    }
 	}
 	break;
+
+#ifdef TLS
+    case TELOPT_START_TLS:
+	if (SB_EOF())
+	    return;
+	if (SB_GET() == TLS_FOLLOWS && !tls_active) {
+	    tls_suspend_iacs = 0;
+	    /* usage: process_rings(netin, netout, netex, ttyin, ttyout, !block) */
+  	    process_rings(1, 0, 0, 0, 0, 1);
+	    if (!tls_try()) {
+		memset(options, 0, sizeof(options));	/* clear any negotiated opt */
+#ifdef COMMENT
+		netringflush();				/* clear net buffers */
+#endif /* COMMENT */
+#if defined(AUTHENTICATION) || defined(ENCRYPTION)
+		auth_encrypt_reset();		/* Forget auth/encrypt state */
+#endif
+		tls_reset_state = 1;		/* flag for state maching reset */
+	    }
+	}
+	return;
+#endif
+
+#ifdef FWD_X
+    case TELOPT_FORWARD_X:
+	if (!fwdx_sbdata || SB_EOF()) /* if fwdx_sbdata == NULL here, something is wrong, this */
+	    return;		      /* suboption is unexpected and we should not continue.   */
+        switch (SB_GET()) {
+	    case FWDX_OPEN: {
+	    	int err, channel;
+		if (SB_LEN() < 2)
+		    break;
+		channel = (subpointer[0] << 8) + subpointer[1];
+		if (err = fwdx_open_channel(channel)) {
+		    unsigned char sb[] = { IAC, SB, TELOPT_FORWARD_X, FWDX_CLOSE, 0, 0, IAC, SE };
+		    sb[4] = subpointer[0]; sb[5] = subpointer[1];
+		    if (sizeof(sb) < NETROOM()) {
+			ring_supply_data(&netoring, sb, sizeof(sb));
+			printsub('>', sb + 2, sizeof(sb) - 2);
+	    	    }
+/*@*/	    	    else printf("telopt_fwdx_close: not enough room in buffer\n");
+		}
+	        break; }
+	    case FWDX_CLOSE: {
+		int channel;
+		if (SB_LEN() < 2)
+		    break;
+		channel = (subpointer[0] << 8) + subpointer[1];
+		fwdx_close_channel(channel);
+	        break; }
+            case FWDX_DATA: {
+                int rc, channel;
+		if (SB_LEN() < 3)
+		    break;
+
+                channel = (subpointer[0] << 8) + subpointer[1];
+                rc = fwdx_send_xauth_to_xserver(channel,subpointer + 2,
+                                                 SB_LEN() - 2);
+                if ( rc >= 0 && SB_LEN()-2-rc > 0 ) {
+                    fwdx_redirect(channel, subpointer + 2, SB_LEN() - 2);
+                }
+	        break; }
+	    case FWDX_OPTIONS:
+    		if (SB_EOF())
+	    	    break;
+	    	fwdx_do_server_options(subpointer, SB_LEN());
+	        break;
+            case FWDX_XOFF: {
+               int channel;
+               if (SB_LEN() < 2)
+                   break;
+               channel = (subpointer[0] << 8) + subpointer[1];
+               fwdx_suspend[channel] = 1;
+               break; }
+            case FWDX_XON: {
+                int channel;
+                if (SB_LEN() < 2)
+                    break;
+                channel = (subpointer[0] << 8) + subpointer[1];
+                fwdx_suspend[channel] = 0;
+                break; }
+            default:
+	        break;
+	}
+    	break;
+#endif  /* FWD_X */
+
     case TELOPT_TSPEED:
 	if (my_want_state_is_wont(TELOPT_TSPEED))
 	    return;
@@ -858,7 +1089,8 @@ suboption()
 
 	    TerminalSpeeds(&ispeed, &ospeed);
 
-	    sprintf((char *)temp, "%c%c%c%c%d,%d%c%c", IAC, SB, TELOPT_TSPEED,
+	    snprintf((char *)temp, sizeof(temp),
+		     "%c%c%c%c%d,%d%c%c", IAC, SB, TELOPT_TSPEED,
 		    TELQUAL_IS, ospeed, ispeed, IAC, SE);
 	    len = strlen((char *)temp+4) + 4;	/* temp[3] is 0 ... */
 
@@ -964,7 +1196,8 @@ suboption()
 		send_wont(TELOPT_XDISPLOC, 1);
 		break;
 	    }
-	    sprintf((char *)temp, "%c%c%c%c%s%c%c", IAC, SB, TELOPT_XDISPLOC,
+	    snprintf((char *)temp, sizeof(temp),
+		     "%c%c%c%c%s%c%c", IAC, SB, TELOPT_XDISPLOC,
 		    TELQUAL_IS, dp, IAC, SE);
 	    len = strlen((char *)temp+4) + 4;	/* temp[3] is 0 ... */
 
@@ -991,6 +1224,13 @@ suboption()
 		case TELQUAL_SEND:
 			if (my_want_state_is_wont(TELOPT_AUTHENTICATION))
 				return;
+#ifdef  TLS	/* TJW: experimental */
+			if(my_state_is_will(TELOPT_START_TLS) &&
+			   !my_want_state_is_wont(TELOPT_START_TLS) &&
+			   !tls_active && tls_enabled) {
+			  break;
+			}
+#endif /* TLS */
 			auth_send(subpointer, SB_LEN());
 			break;
 		case TELQUAL_REPLY:
@@ -1075,7 +1315,7 @@ suboption()
 
 static unsigned char str_lm[] = { IAC, SB, TELOPT_LINEMODE, 0, 0, IAC, SE };
 
-    void
+void
 lm_will(cmd, len)
     unsigned char *cmd;
     int len;
@@ -1098,7 +1338,7 @@ lm_will(cmd, len)
     }
 }
 
-    void
+void
 lm_wont(cmd, len)
     unsigned char *cmd;
     int len;
@@ -1115,7 +1355,7 @@ lm_wont(cmd, len)
     }
 }
 
-    void
+void
 lm_do(cmd, len)
     unsigned char *cmd;
     int len;
@@ -1138,7 +1378,7 @@ lm_do(cmd, len)
     }
 }
 
-    void
+void
 lm_dont(cmd, len)
     unsigned char *cmd;
     int len;
@@ -1159,7 +1399,7 @@ static unsigned char str_lm_mode[] = {
 	IAC, SB, TELOPT_LINEMODE, LM_MODE, 0, IAC, SE
 };
 
-	void
+void
 lm_mode(cmd, len, init)
 	unsigned char *cmd;
 	int len, init;
@@ -1201,7 +1441,7 @@ struct spc {
 #define SLC_RVALUE	2
 static int slc_mode = SLC_EXPORT;
 
-	void
+void
 slc_init()
 {
 	register struct spc *spcp;
@@ -1263,7 +1503,7 @@ slc_init()
 
 }
 
-    void
+void
 slcstate()
 {
     printf("Special characters are %s values\n",
@@ -1272,7 +1512,7 @@ slcstate()
 					 "remote");
 }
 
-    void
+void
 slc_mode_export()
 {
     slc_mode = SLC_EXPORT;
@@ -1280,7 +1520,7 @@ slc_mode_export()
 	slc_export();
 }
 
-    void
+void
 slc_mode_import(def)
     int def;
 {
@@ -1296,7 +1536,7 @@ unsigned char slc_import_def[] = {
 	IAC, SB, TELOPT_LINEMODE, LM_SLC, 0, SLC_DEFAULT, 0, IAC, SE
 };
 
-    void
+void
 slc_import(def)
     int def;
 {
@@ -1312,7 +1552,7 @@ slc_import(def)
 /*@*/ else printf("slc_import: not enough room\n");
 }
 
-    void
+void
 slc_export()
 {
     register struct spc *spcp;
@@ -1336,7 +1576,7 @@ slc_export()
     setconnmode(1);	/* Make sure the character values are set */
 }
 
-	void
+void
 slc(cp, len)
 	register unsigned char *cp;
 	int len;
@@ -1406,7 +1646,7 @@ slc(cp, len)
 		setconnmode(1);	/* set the  new character values */
 }
 
-    void
+void
 slc_check()
 {
     register struct spc *spcp;
@@ -1430,7 +1670,7 @@ slc_check()
 unsigned char slc_reply[128];
 unsigned char *slc_replyp;
 
-	void
+void
 slc_start_reply()
 {
 	slc_replyp = slc_reply;
@@ -1440,7 +1680,7 @@ slc_start_reply()
 	*slc_replyp++ = LM_SLC;
 }
 
-	void
+void
 slc_add_reply(func, flags, value)
 	unsigned char func;
 	unsigned char flags;
@@ -1454,7 +1694,7 @@ slc_add_reply(func, flags, value)
 		*slc_replyp++ = IAC;
 }
 
-    void
+void
 slc_end_reply()
 {
     register int len;
@@ -1471,7 +1711,7 @@ slc_end_reply()
 /*@*/else printf("slc_end_reply: not enough room\n");
 }
 
-	int
+int
 slc_update()
 {
 	register struct spc *spcp;
@@ -1509,7 +1749,7 @@ int old_env_value = OLD_ENV_VALUE;
 # endif
 #endif
 
-	void
+void
 env_opt(buf, len)
 	register unsigned char *buf;
 	register int len;
@@ -1583,12 +1823,17 @@ unsigned char *opt_reply;
 unsigned char *opt_replyp;
 unsigned char *opt_replyend;
 
-	void
+void
 env_opt_start()
 {
-	if (opt_reply)
-		opt_reply = (unsigned char *)realloc(opt_reply, OPT_REPLY_SIZE);
-	else
+    if (opt_reply) {
+                unsigned char * p;
+                p = (unsigned char *)realloc(opt_reply, OPT_REPLY_SIZE);
+                if (p == NULL) {
+                    free(opt_reply);
+                    opt_reply = NULL;
+                }
+    } else
 		opt_reply = (unsigned char *)malloc(OPT_REPLY_SIZE);
 	if (opt_reply == NULL) {
 /*@*/		printf("env_opt_start: malloc()/realloc() failed!!!\n");
@@ -1603,7 +1848,7 @@ env_opt_start()
 	*opt_replyp++ = TELQUAL_IS;
 }
 
-	void
+void
 env_opt_start_info()
 {
 	env_opt_start();
@@ -1611,7 +1856,7 @@ env_opt_start_info()
 	    opt_replyp[-1] = TELQUAL_INFO;
 }
 
-	void
+void
 env_opt_add(ep)
 	register unsigned char *ep;
 {
@@ -1637,9 +1882,14 @@ env_opt_add(ep)
 				strlen((char *)ep) + 6 > opt_replyend)
 	{
 		register int len;
+                unsigned char *p;
 		opt_replyend += OPT_REPLY_SIZE;
 		len = opt_replyend - opt_reply;
-		opt_reply = (unsigned char *)realloc(opt_reply, len);
+		p = (unsigned char *)realloc(opt_reply, len);
+                if (p == NULL) {
+                    free(opt_reply);
+                    opt_reply = NULL;
+                }
 		if (opt_reply == NULL) {
 /*@*/			printf("env_opt_add: realloc() failed!!!\n");
 			opt_reply = opt_replyp = opt_replyend = NULL;
@@ -1685,7 +1935,7 @@ env_opt_add(ep)
 	}
 }
 
-	int
+int
 opt_welldefined(ep)
 	char *ep;
 {
@@ -1698,7 +1948,8 @@ opt_welldefined(ep)
 		return(1);
 	return(0);
 }
-	void
+
+void
 env_opt_end(emptyok)
 	register int emptyok;
 {
@@ -1722,7 +1973,7 @@ env_opt_end(emptyok)
 
 
 
-    int
+int
 telrcv()
 {
     register int c;
@@ -1730,6 +1981,9 @@ telrcv()
     register unsigned char *sbp;
     int count;
     int returnValue = 0;
+#ifdef FWD_X
+    unsigned char *subroof;
+#endif
 
     scc = 0;
     count = 0;
@@ -1791,7 +2045,7 @@ telrcv()
 #	    endif /* defined(TN3270) */
 		    /*
 		     * The 'crmod' hack (see following) is needed
-		     * since we can't * set CRMOD on output only.
+		     * since we can't set CRMOD on output only.
 		     * Machines like MULTICS like to send \r without
 		     * \n; since we must turn off CRMOD to get proper
 		     * input, the mapping is done here (sigh).
@@ -1949,6 +2203,17 @@ process_iac:
 	    if (c == IAC) {
 		telrcv_state = TS_SE;
 	    } else {
+#ifdef FWD_X
+		/* flush the subbuffer if it is going to overflow */
+		subroof = subbuffer + sizeof(subbuffer) - 3;
+		if ((subpointer >= subroof) && subbuffer[0] == TELOPT_FORWARD_X && 
+			subbuffer[1] == FWDX_DATA) {
+		    int channel = (subbuffer[2] << 8) + subbuffer[3];
+	    	    fwdx_redirect(channel, subbuffer + 4, subpointer - subbuffer - 4);
+		    /* start filling subbuffer again after the "channel" bytes */
+		    subpointer = subbuffer + 4;
+		}
+#endif
 		SB_ACCUM(c);
 	    }
 	    continue;
@@ -1979,6 +2244,17 @@ process_iac:
 		    telrcv_state = TS_IAC;
 		    goto process_iac;
 		}
+#ifdef FWD_X
+		/* flush the subbuffer if it is going to overflow */
+		subroof = subbuffer + sizeof(subbuffer) - 3;
+		if ((subpointer >= subroof) && (subbuffer[0] == TELOPT_FORWARD_X) &&
+			(subbuffer[1] == FWDX_DATA)) {
+		    int channel = (subbuffer[2] << 8) + subbuffer[3];
+	    	    fwdx_redirect(channel, subbuffer + 4, subpointer - subbuffer - 4);
+		    /* start filling subbuffer again after the "channel" bytes */
+		    subpointer = subbuffer + 4;
+		}
+#endif
 		SB_ACCUM(c);
 		telrcv_state = TS_SB;
 	    } else {
@@ -1987,6 +2263,17 @@ process_iac:
 		subpointer -= 2;
 		SB_TERM();
 		suboption();	/* handle sub-option */
+#ifdef TLS
+		/* if we just negotiated TLS, we must reset the telnet state machine.
+		 * we do this by simply abort this processing.
+		 */
+		if (tls_reset_state) {
+		    tls_reset_state = 0;
+                    reset_telnet();
+                    announce_my_options();
+		    return returnValue||count;
+		}
+#endif /* TLS */
 		SetIn3270();
 		telrcv_state = TS_DATA;
 	    }
@@ -1999,7 +2286,7 @@ process_iac:
 
 static int bol = 1, local = 0;
 
-    int
+int
 rlogin_susp()
 {
     if (local) {
@@ -2011,7 +2298,7 @@ rlogin_susp()
     return(0);
 }
 
-    static int
+static int
 telsnd()
 {
     int tcc;
@@ -2162,7 +2449,7 @@ telsnd()
  */
 
 
-    int
+int
 Scheduler(block)
     int	block;			/* should we block in the select ? */
 {
@@ -2208,6 +2495,9 @@ Scheduler(block)
     }
 #endif	/* defined(TN3270) && defined(unix) */
 
+    if (scheduler_lockout_tty)
+        ttyin = ttyout = 0;
+
     /* Call to system code to process rings */
 
     returnValue = process_rings(netin, netout, netex, ttyin, ttyout, !block);
@@ -2236,126 +2526,184 @@ Scheduler(block)
     if (ring_full_count(&netiring)) {
 #	if !defined(TN3270)
 	returnValue |= telrcv();
+#	if defined(TLS)
+	/* having a while() loop here, as in the telnetd code, can be dangerous
+	 * if we are receiving continueous data from the server (for example if
+	 * you are using FORWARD_X and are using Netscape to look at a page with
+	 * animated GIF's), we might get stuck here in an endless loop.
+	 */
+	if (tls_pending() > 0) {
+	    int c, canread;
+
+	    canread = ring_empty_consecutive(&netiring);
+	    c = RECV(net, (char *)netiring.supply, canread, 0);
+	    if (c < 0 && (errno == EWOULDBLOCK || errno == EINTR))
+	    	c = 0;
+	    if (netdata) {
+	    	Dump('<', netiring.supply, c);
+	    }
+	    if (c > 0) {
+	    	ring_supplied(&netiring, c);
+	    	returnValue |= telrcv();
+	    }
+	}
+#	endif /* defined(TLS) */
 #	else /* !defined(TN3270) */
 	returnValue = Push3270();
 #	endif /* !defined(TN3270) */
     }
     return returnValue;
 }
+
+void announce_my_options(void)
+{
+#ifdef ENCRYPTION
+    int printed_encrypt = 0;
+#endif /* ENCRYPTION */
+
+#ifdef TLS
+	if(!tls_active && tls_enabled)
+		send_will(TELOPT_START_TLS, 1);
+                while ( his_do_dont_is_changing(TELOPT_START_TLS) )
+                    telnet_spin();
+#endif
+#if	defined(AUTHENTICATION)
+    if (autologin) {
+		send_will(TELOPT_AUTHENTICATION, 1);
+                while ( his_do_dont_is_changing(TELOPT_AUTHENTICATION) )
+                    telnet_spin();
+                if (his_state_is_do(TELOPT_AUTHENTICATION))
+                    auth_wait(NULL);
+    }
+#endif
+#ifdef	ENCRYPTION
+#  ifdef TLS
+	if (!tls_active) {	/* Don't use TELOPT ENCRYPT under TLS */
+#  endif
+            if (his_state_is_do(TELOPT_AUTHENTICATION) &&
+                 wantencryption || auth_must_encrypt()) {
+                time_t timeout = time(0) + 60;
+	send_do(TELOPT_ENCRYPT, 1);
+	send_will(TELOPT_ENCRYPT, 1);
+
+                /*
+                 * Note: we assume a tie to the authentication option here.  This
+                 * is necessary so that if authentication fails, we don't spin
+                 * forever. 
+                 */
+
+                while (1) {
+#ifdef AUTHENTICATION
+                    extern int auth_has_failed;
+
+                    if (my_want_state_is_wont(TELOPT_AUTHENTICATION)) {
+                        printf("\r\nServer refused to negotiate authentication, which is required\r\n");
+                        printf("for encryption.  Good bye.\r\n");
+                        Exit(1);
+                    }
+                    if (auth_has_failed) {
+                        printf("\r\nAuthentication negotation has failed, which is required for\r\n");
+                        printf("encryption.  Good bye.\r\n");
+                        Exit(1);
+                    }
+#endif /* AUTHENTICATION */
+                    if (my_want_state_is_dont(TELOPT_ENCRYPT) ||
+                         my_want_state_is_wont(TELOPT_ENCRYPT)) {
+                        printf("\r\nServer refused to negotiate encryption.  Good bye.\r\n");
+                        Exit(1);
+                    }
+                    if (encrypt_is_encrypting())
+                        break;
+                    if (time(0) > timeout) {
+                        printf("\r\nEncryption could not be enabled.  Goodbye.\r\n");
+                        Exit(1);
+                    }
+                    if (printed_encrypt == 0) {
+                        printed_encrypt = 1;
+                        printf("Waiting for encryption to be negotiated...\r\n");
+                        /*
+                        * Turn on MODE_TRAPSIG and then turn off localchars 
+                        * so that ^C will cause telnet to exit.
+                        */
+                        TerminalNewMode(getconnmode()|MODE_TRAPSIG);
+                        intr_waiting = 1;
+                    }
+                    if (intr_happened) {
+                        printf("\r\nUser requested an interrupt.  Goodbye.\r\n");
+                        Exit(1);
+                    }
+                    telnet_spin();
+                }
+                if (printed_encrypt) {
+                    printf("done.\r\n");
+                    intr_waiting = 0;
+                    setconnmode(0);
+                }
+            }
+#  ifdef TLS
+	}
+#  endif
+#endif	/* ENCRYPTION */
+#ifdef FWD_X
+        if (fwdx_enable_flag && env_getvalue((unsigned char *)"DISPLAY")
+#ifdef ENCRYPTION
+             && encrypt_is_encrypting()
+#endif /* ENCRYPTION */
+             ) 
+        {
+	    if (!fwdx_sbdata)
+	    	fwdx_sbdata = malloc(fwdx_sbdata_size);
+	    if (fwdx_sbdata)
+	    	send_do(TELOPT_FORWARD_X, 1);
+	}
+#endif  /* FWD_X */
+	send_do(TELOPT_SGA, 1);
+	send_will(TELOPT_TTYPE, 1);
+	send_will(TELOPT_NAWS, 1);
+	send_will(TELOPT_TSPEED, 1);
+	send_will(TELOPT_LFLOW, 1);
+	send_will(TELOPT_LINEMODE, 1);
+	send_will(TELOPT_NEW_ENVIRON, 1);
+	send_do(TELOPT_STATUS, 1);
+	if (env_getvalue((unsigned char *)"DISPLAY") 
+#ifdef FWD_X
+             && !fwdx_sbdata
+#endif /* FWD_X */
+             )
+	    send_will(TELOPT_XDISPLOC, 1);
+	if (eight)
+	    tel_enter_binary(eight);
+}
+
 
 /*
  * Select from tty and network...
  */
     void
-telnet(user)
+telnet(user, explicit)
     char *user;
+    int explicit;
 {
-    int printed_encrypt = 0;
-
     sys_telnet_init();
 
 #if	defined(AUTHENTICATION) || defined(ENCRYPTION)
     {
-	static char local_host[256] = { 0 };
+	extern char *localhost ();
+	static char *local_host = 0;
+
+	if (!local_host)
+		local_host = localhost ();
 
-	if (!local_host[0]) {
-		gethostname(local_host, sizeof(local_host));
-		local_host[sizeof(local_host)-1] = 0;
-	}
 	auth_encrypt_init(local_host, hostname, "TELNET", 0);
-	auth_encrypt_user(user);
+	auth_encrypt_user(user, explicit);
     }
 #endif	/* defined(AUTHENTICATION) || defined(ENCRYPTION)  */
-    if (telnetport) {
 #   if !defined(TN3270)
-#if	defined(AUTHENTICATION)
-	if (autologin)
-		send_will(TELOPT_AUTHENTICATION, 1);
-#endif
-#ifdef	ENCRYPTION
-        if (telnetport || wantencryption) {
-            send_do(TELOPT_ENCRYPT, 1);
-            send_will(TELOPT_ENCRYPT, 1);
-        }
-#endif	/* ENCRYPTION */
-        if (telnetport) {
-            send_do(TELOPT_SGA, 1);
-            send_will(TELOPT_TTYPE, 1);
-            send_will(TELOPT_NAWS, 1);
-            send_will(TELOPT_TSPEED, 1);
-            send_will(TELOPT_LFLOW, 1);
-            send_will(TELOPT_LINEMODE, 1);
-            send_will(TELOPT_NEW_ENVIRON, 1);
-            send_do(TELOPT_STATUS, 1);
-            if (env_getvalue((unsigned char *)"DISPLAY") &&
-                env_is_exported((unsigned char *)"DISPLAY"))
-                send_will(TELOPT_XDISPLOC, 1);
-            if (eight)
-                tel_enter_binary(eight);
-        }
+    if (telnetport) {
+    	announce_my_options();
+    }
 #   endif /* !defined(TN3270) */
 
-#if 0 /* TJW: This doesn't work!  telnet_spin() is a stub in this version! */
-#ifdef ENCRYPTION
-    /*
-     * Note: we assume a tie to the authentication option here.  This
-     * is necessary so that authentication fails, we don't spin
-     * forever. 
-     */
-    if (wantencryption) {
-	extern int auth_has_failed;
-	time_t timeout = time(0) + 60;
-
-	send_do(TELOPT_ENCRYPT, 1);
-	send_will(TELOPT_ENCRYPT, 1);
-	while (1) {
-	    if (my_want_state_is_wont(TELOPT_AUTHENTICATION)) {
-		printf("\nServer refused to negotiation authentication, which is required\n");
-		printf("for encryption.  Good bye.\n\r");
-		Exit(1);
-	    }
-	    if (auth_has_failed) {
-		printf("\nAuthentication negotation has failed, which is required for\n");
-		printf("encryption.  Good bye.\n\r");
-		Exit(1);
-	    }
-	    if (my_want_state_is_dont(TELOPT_ENCRYPT) ||
-		my_want_state_is_wont(TELOPT_ENCRYPT)) {
-		printf("\nServer refused to negotiate encryption.  Good bye.\n\r");
-		Exit(1);
-	    }
-	    if (encrypt_is_encrypting())
-		break;
-	    if (time(0) > timeout) {
-		printf("\nEncryption could not be enabled.  Goodbye.\n\r");
-		Exit(1);
-	    }
-	    if (printed_encrypt == 0) {
-		    printed_encrypt = 1;
-		    printf("Waiting for encryption to be negotiated...\n");
-		    /*
-		     * Turn on MODE_TRAPSIG and then turn off localchars 
-		     * so that ^C will cause telnet to exit.
-		     */
-		    TerminalNewMode(getconnmode()|MODE_TRAPSIG);
-		    intr_waiting = 1;
-	    }
-	    if (intr_happened) {
-		    printf("\nUser requested an interrupt.  Goodbye.\n\r");
-		    Exit(1);
-	    }
-	    telnet_spin();
-	}
-	if (printed_encrypt) {
-		printf("done.\n");
-		intr_waiting = 0;
-		setconnmode(0);
-	}
-    }
-#endif
-#endif
-    }
-
 #   if !defined(TN3270)
     for (;;) {
 	int schedValue;
@@ -2425,7 +2773,7 @@ telnet(user)
  * character.
  */
 
-    static char *
+static char *
 nextitem(current)
     char *current;
 {
@@ -2473,7 +2821,7 @@ nextitem(current)
  * us in any case.
  */
 
-    static void
+static void
 netclear()
 {
 #if	0	/* XXX */
@@ -2516,7 +2864,7 @@ netclear()
  * These routines add various telnet commands to the data stream.
  */
 
-    static void
+static void
 doflush()
 {
     NET2ADD(IAC, DO);
@@ -2528,7 +2876,7 @@ doflush()
     printoption("SENT", DO, TELOPT_TM);
 }
 
-    void
+void
 xmitAO()
 {
     NET2ADD(IAC, AO);
@@ -2539,14 +2887,14 @@ xmitAO()
 }
 
 
-    void
+void
 xmitEL()
 {
     NET2ADD(IAC, EL);
     printoption("SENT", IAC, EL);
 }
 
-    void
+void
 xmitEC()
 {
     NET2ADD(IAC, EC);
@@ -2554,7 +2902,7 @@ xmitEC()
 }
 
 
-    int
+int
 dosynch()
 {
     netclear();			/* clear the path to the network */
@@ -2567,7 +2915,7 @@ dosynch()
 
 int want_status_response = 0;
 
-    int
+int
 get_status()
 {
     unsigned char tmp[16];
@@ -2593,7 +2941,7 @@ get_status()
     return 1;
 }
 
-    void
+void
 intp()
 {
     NET2ADD(IAC, IP);
@@ -2607,7 +2955,7 @@ intp()
     }
 }
 
-    void
+void
 sendbrk()
 {
     NET2ADD(IAC, BREAK);
@@ -2621,7 +2969,7 @@ sendbrk()
     }
 }
 
-    void
+void
 sendabort()
 {
     NET2ADD(IAC, ABORT);
@@ -2635,7 +2983,7 @@ sendabort()
     }
 }
 
-    void
+void
 sendsusp()
 {
     NET2ADD(IAC, SUSP);
@@ -2649,14 +2997,14 @@ sendsusp()
     }
 }
 
-    void
+void
 sendeof()
 {
     NET2ADD(IAC, xEOF);
     printoption("SENT", IAC, xEOF);
 }
 
-    void
+void
 sendayt()
 {
     NET2ADD(IAC, AYT);
@@ -2667,7 +3015,7 @@ sendayt()
  * Send a window size update to the remote system.
  */
 
-    void
+void
 sendnaws()
 {
     long rows, cols;
@@ -2699,7 +3047,7 @@ sendnaws()
     }
 }
 
-    void
+void
 tel_enter_binary(rw)
     int rw;
 {
@@ -2709,7 +3057,7 @@ tel_enter_binary(rw)
 	send_will(TELOPT_BINARY, 1);
 }
 
-    void
+void
 tel_leave_binary(rw)
     int rw;
 {
