@@ -68,9 +68,9 @@ static char copyright[] =
 # endif /* HAVE_SYS_UN_H  */
 #endif  /* FWDX_UNIX_SOCK */
 
-#ifdef SRP
+#ifdef HAVE_SRP
 #include <t_pwd.h>
-#endif /* SRP */
+#endif /* HAVE_SRP */
 
 #ifdef KRB5
 #include <krb5.h>
@@ -285,7 +285,7 @@ unsigned int fwdx_random(void)
     unsigned int rv;
     FILE *rf;
     
-#ifdef SRP
+#ifdef HAVE_SRP
     t_random((unsigned char *) &rv, sizeof(unsigned int));
     return rv;
 #elif defined(TLS)
