@@ -1026,8 +1026,14 @@ dooption(option)
 			if (!fwdx_disable_flag) {
 				if (!fwdx_sbdata)
 					fwdx_sbdata = malloc(fwdx_sbdata_size);
+# ifdef ENCRYPTION
 				if (fwdx_sbdata && encrypt_is_encrypting())
 					changeok++;
+# endif /* ENCRYPTION */
+# ifdef TLS
+				if (fwdx_sbdata && tls_active)
+					changeok++;
+# endif /* TLS */
 			}
 			break;
 #endif	/* FWD_X */
