@@ -30,30 +30,75 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  */
+/*
+ * START_TLS and FORWARD_X support from Peter Runestig <peter@runestig.com>
+ * and Jeffrey Altman <jaltman@columbia.edu> incorporated into the
+ * SRP Telnet distribution 10/19/2000 by Tom Wu <tjw@cs.stanford.edu>.
+ */
 
 #ifndef lint
 static char sccsid[] = "@(#)sys_term.c	8.4+1 (Berkeley) 5/30/95";
 #endif /* not lint */
 
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
+
+#ifdef TLS
+#define WRITE tls_write
+#else
+#define WRITE write
+#endif
+
 #include "telnetd.h"
 #include "pathnames.h"
 
+#ifndef LOGIN_PROGRAM
+#define LOGIN_PROGRAM _PATH_LOGIN
+#endif
+
 #if	defined(AUTHENTICATION)
 #include <libtelnet/auth.h>
 #endif
 
+#include <pwd.h>
+
 #if defined(CRAY) || defined(__hpux)
 # define PARENT_DOES_UTMP
 #endif
 
+char *login_program = LOGIN_PROGRAM;
+
 #ifdef	NEWINIT
 #include <initreq.h>
 int	utmp_len = MAXHOSTNAMELEN;	/* sizeof(init_request.host) */
 #else	/* NEWINIT*/
 
+/* Don't use UTMPX with HP-UX */
+# if defined(__hpux) && defined(UTMPX)
+#  undef UTMPX
+# endif
+
+/* Some OSes, like AIX, use UTMPX but don't have wtmpx */
+# if defined(_AIX) && !defined(NO_WTMPX)
+#  define NO_WTMPX 1
+# endif /* _AIX */
+
 # ifdef	UTMPX
 # include <utmpx.h>
+# ifdef NO_WTMPX
+#  include <utmp.h>
+struct	utmp wtmp;
+# else
 struct	utmpx wtmp;
+# endif
+# ifndef WTMPX_FILE
+#  ifdef _PATH_WTMPX
+#   define WTMPX_FILE _PATH_WTMPX
+#  elif defined(linux) && !defined(__USE_GNU)
+#   define WTMPX_FILE "/var/log/wtmpx"
+#  endif /* _PATH_WTMPX */
+# endif /* WTMPX_FILE */
 # else
 # include <utmp.h>
 struct	utmp wtmp;
@@ -96,8 +141,8 @@ extern struct sysv sysv;
 #endif	/* NEWINIT */
 
 #ifdef	STREAMSPTY
-# ifndef OSF
-#include <sac.h>
+# ifdef HAVE_SAC_H
+#  include <sac.h>
 # endif
 #include <sys/stropts.h>
 #endif
@@ -105,16 +150,13 @@ extern struct sysv sysv;
 #define SCPYN(a, b)	(void) strncpy(a, b, sizeof(a))
 #define SCMPN(a, b)	strncmp(a, b, sizeof(a))
 
-#ifdef	STREAMS
+#ifdef	HAVE_SYS_STREAM_H
 #include <sys/stream.h>
 #endif
 #ifdef __hpux
 #include <sys/resource.h>
 #include <sys/proc.h>
 #endif
-#if !defined(IRIX) && !defined(LINUX)
-#include <sys/tty.h>
-#endif
 #ifdef	t_erase
 #undef	t_erase
 #undef	t_kill
@@ -199,7 +241,7 @@ int ttyfd = -1;
  * set_termbuf() writes the structure into the kernel.
  */
 
-	void
+void
 init_termbuf()
 {
 #ifndef	USE_TERMIO
@@ -220,7 +262,7 @@ init_termbuf()
 }
 
 #if	defined(LINEMODE) && defined(TIOCPKT_IOCTL)
-	void
+void
 copy_termbuf(cp, len)
 	char *cp;
 	int len;
@@ -232,7 +274,7 @@ copy_termbuf(cp, len)
 }
 #endif	/* defined(LINEMODE) && defined(TIOCPKT_IOCTL) */
 
-	void
+void
 set_termbuf()
 {
 	/*
@@ -276,7 +318,7 @@ set_termbuf()
  */
 
 #ifndef	USE_TERMIO
-	int
+int
 spcset(func, valp, valpp)
 	int func;
 	cc_t *valp;
@@ -351,7 +393,7 @@ spcset(func, valp, valpp)
 
 #else	/* USE_TERMIO */
 
-	int
+int
 spcset(func, valp, valpp)
 	int func;
 	cc_t *valp;
@@ -453,7 +495,7 @@ spcset(func, valp, valpp)
  *
  * Return the number of pty's configured into the system.
  */
-	int
+int
 getnpty()
 {
 #ifdef _SC_CRAY_NPTY
@@ -486,16 +528,28 @@ char *line = Xline;
 char *myline = "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0";
 #endif	/* CRAY */
 
-	int
+int
 getpty(ptynum)
 int *ptynum;
 {
-	register int p;
+	register int p = -1;
 #ifdef	STREAMSPTY
 	int t;
 	char *ptsname();
 
-	p = open("/dev/ptmx", 2);
+#ifdef __hpux
+	if(p < 0)
+	  p = open("/dev/ptym/clone", O_RDWR|O_NDELAY);
+#endif
+	if(p < 0)
+	  p = open("/dev/ptmx", O_RDWR|O_NDELAY);
+#ifdef _AIX
+	if(p < 0)
+	  p = open("/dev/ptc", O_RDWR|O_NDELAY);
+#endif
+	if(p < 0)
+	  p = open("/dev/pty", O_RDWR|O_NDELAY);
+
 	if (p > 0) {
 		grantpt(p);
 		unlockpt(p);
@@ -533,9 +587,14 @@ int *ptynum;
 		 */
 		if (stat(line, &stb) < 0)
 			break;
+#ifdef SCO
+		for (i = 0; i < 9999; i++) {
+			sprintf(p2, "%d", i);
+#else
 		for (i = 0; i < 16; i++) {
 			*p2 = "0123456789abcdef"[i];
-			p = open(line, 2);
+#endif /* SCO */
+			p = open(line, O_RDWR);
 			if (p > 0) {
 #ifndef	__hpux
 				line[5] = 't';
@@ -564,7 +623,7 @@ int *ptynum;
 
 	for (*ptynum = lowpty; *ptynum <= highpty; (*ptynum)++) {
 		(void) sprintf(myline, "/dev/pty/%03d", *ptynum);
-		p = open(myline, 2);
+		p = open(myline, O_RDWR);
 		if (p < 0)
 			continue;
 		(void) sprintf(line, "/dev/ttyp%03d", *ptynum);
@@ -580,7 +639,7 @@ int *ptynum;
 			chown(line, 0, 0);
 			chmod(line, 0600);
 			(void)close(p);
-			p = open(myline, 2);
+			p = open(myline, O_RDWR);
 			if (p < 0)
 				continue;
 		}
@@ -626,7 +685,7 @@ int *ptynum;
 static int linestate;
 #endif
 
-	int
+int
 tty_linemode()
 {
 #ifndef convex
@@ -640,7 +699,7 @@ tty_linemode()
 #endif
 }
 
-	void
+void
 tty_setlinemode(on)
 	int on;
 {
@@ -665,7 +724,7 @@ tty_setlinemode(on)
 }
 #endif	/* LINEMODE */
 
-	int
+int
 tty_isecho()
 {
 #ifndef USE_TERMIO
@@ -675,7 +734,7 @@ tty_isecho()
 #endif
 }
 
-	int
+int
 tty_flowmode()
 {
 #ifndef USE_TERMIO
@@ -685,7 +744,7 @@ tty_flowmode()
 #endif
 }
 
-	int
+int
 tty_restartany()
 {
 #ifndef USE_TERMIO
@@ -699,7 +758,7 @@ tty_restartany()
 #endif
 }
 
-	void
+void
 tty_setecho(on)
 	int on;
 {
@@ -716,7 +775,7 @@ tty_setecho(on)
 #endif
 }
 
-	int
+int
 tty_israw()
 {
 #ifndef USE_TERMIO
@@ -727,7 +786,7 @@ tty_israw()
 }
 
 #if	defined (AUTHENTICATION) && defined(NO_LOGIN_F) && defined(LOGIN_R)
-	int
+int
 tty_setraw(on)
 {
 #  ifndef USE_TERMIO
@@ -744,7 +803,7 @@ tty_setraw(on)
 }
 #endif
 
-	void
+void
 tty_binaryin(on)
 	int on;
 {
@@ -762,7 +821,7 @@ tty_binaryin(on)
 #endif
 }
 
-	void
+void
 tty_binaryout(on)
 	int on;
 {
@@ -784,7 +843,7 @@ tty_binaryout(on)
 #endif
 }
 
-	int
+int
 tty_isbinaryin()
 {
 #ifndef	USE_TERMIO
@@ -794,7 +853,7 @@ tty_isbinaryin()
 #endif
 }
 
-	int
+int
 tty_isbinaryout()
 {
 #ifndef	USE_TERMIO
@@ -805,7 +864,7 @@ tty_isbinaryout()
 }
 
 #ifdef	LINEMODE
-	int
+int
 tty_isediting()
 {
 #ifndef USE_TERMIO
@@ -815,7 +874,7 @@ tty_isediting()
 #endif
 }
 
-	int
+int
 tty_istrapsig()
 {
 #ifndef USE_TERMIO
@@ -825,7 +884,7 @@ tty_istrapsig()
 #endif
 }
 
-	void
+void
 tty_setedit(on)
 	int on;
 {
@@ -842,7 +901,7 @@ tty_setedit(on)
 #endif
 }
 
-	void
+void
 tty_setsig(on)
 	int on;
 {
@@ -858,7 +917,7 @@ tty_setsig(on)
 }
 #endif	/* LINEMODE */
 
-	int
+int
 tty_issofttab()
 {
 #ifndef	USE_TERMIO
@@ -873,7 +932,7 @@ tty_issofttab()
 #endif
 }
 
-	void
+void
 tty_setsofttab(on)
 	int on;
 {
@@ -903,7 +962,7 @@ tty_setsofttab(on)
 #endif
 }
 
-	int
+int
 tty_islitecho()
 {
 #ifndef	USE_TERMIO
@@ -921,7 +980,7 @@ tty_islitecho()
 #endif
 }
 
-	void
+void
 tty_setlitecho(on)
 	int on;
 {
@@ -946,7 +1005,7 @@ tty_setlitecho(on)
 #endif
 }
 
-	int
+int
 tty_iscrnl()
 {
 #ifndef	USE_TERMIO
@@ -1006,7 +1065,7 @@ struct termspeeds {
 };
 #endif	/* DECODE_BUAD */
 
-	void
+void
 tty_tspeed(val)
 	int val;
 {
@@ -1023,7 +1082,7 @@ tty_tspeed(val)
 #endif	/* DECODE_BUAD */
 }
 
-	void
+void
 tty_rspeed(val)
 	int val;
 {
@@ -1041,7 +1100,7 @@ tty_rspeed(val)
 }
 
 #if	defined(CRAY2) && defined(UNICOS5)
-	int
+int
 tty_isnewmap()
 {
 	return((termbuf.c_oflag & OPOST) && (termbuf.c_oflag & ONLCR) &&
@@ -1056,7 +1115,7 @@ extern char wtmpf[];
 int	gotalarm;
 
 	/* ARGSUSED */
-	void
+void
 nologinproc(sig)
 	int sig;
 {
@@ -1081,7 +1140,7 @@ extern void utmp_sig_notify P((int));
  * that is necessary.  The return value is a file descriptor
  * for the slave side.
  */
-	int
+int
 getptyslave()
 {
 	register int t = -1;
@@ -1136,15 +1195,21 @@ getptyslave()
 #ifdef	USE_TERMIO
 	ttyfd = t;
 #endif
+#ifndef _AIX
 	if (ioctl(t, I_PUSH, "ptem") < 0)
 		fatal(net, "I_PUSH ptem");
 	if (ioctl(t, I_PUSH, "ldterm") < 0)
 		fatal(net, "I_PUSH ldterm");
+#endif
 #if !defined(IRIX) && !defined(OSF)
+# ifdef _AIX
+	ioctl(pty, I_PUSH, "pckt");
+# else
 	if (ioctl(t, I_PUSH, "ttcompat") < 0)
 		fatal(net, "I_PUSH ttcompat");
 	if (ioctl(pty, I_PUSH, "pckt") < 0)
 		fatal(net, "I_PUSH pckt");
+# endif
 #endif
 #endif
 
@@ -1171,7 +1236,7 @@ getptyslave()
 	/*
 	 * Settings for UNICOS (and HPUX)
 	 */
-# if defined(CRAY) || defined(__hpux)
+# if defined(CRAY) || defined(__hpux) || defined(SCO)
 	termbuf.c_oflag = OPOST|ONLCR|TAB3;
 	termbuf.c_iflag = IGNPAR|ISTRIP|ICRNL|IXON;
 	termbuf.c_lflag = ISIG|ICANON|ECHO|ECHOE|ECHOK;
@@ -1229,7 +1294,7 @@ getptyslave()
  * Open the specified slave side of the pty,
  * making sure that we have a clean tty.
  */
-	int
+int
 cleanopen(line)
 	char *line;
 {
@@ -1281,7 +1346,9 @@ cleanopen(line)
 	 */
 # if !(defined(CRAY) || defined(__hpux)) && (BSD <= 43) && !defined(STREAMSPTY)
 	(void) signal(SIGHUP, SIG_IGN);
+#ifdef HAVE_VHANGUP
 	vhangup();
+#endif /* HAVE_VHANGUP */
 	(void) signal(SIGHUP, SIG_DFL);
 	t = open(line, O_RDWR|O_NOCTTY);
 	if (t < 0)
@@ -1328,10 +1395,11 @@ cleanopen(line)
 
 #if BSD <= 43
 
-	int
+int
 login_tty(t)
 	int t;
 {
+#ifndef _AIX	/* Is all this really needed? */
 	if (setsid() < 0) {
 #ifdef ultrix
 		/*
@@ -1363,13 +1431,14 @@ login_tty(t)
 	 * setsid() call above may have set our pgrp, so clear
 	 * it out before opening the tty...
 	 */
-#  if !defined(SOLARIS) && !defined(IRIX) && !defined(__hpux)
-	(void) setpgrp(0, 0);
-#  else
+#  ifdef SETPGRP_VOID
 	(void) setpgrp();
-#  endif
+#  else
+	(void) setpgrp(0, 0);
+#  endif /* SETPGRP_VOID */
 	close(open(line, O_RDWR));
 # endif
+#endif /* _AIX */
 	if (t != 0)
 		(void) dup2(t, 0);
 	if (t != 1)
@@ -1382,6 +1451,8 @@ login_tty(t)
 }
 #endif	/* BSD <= 43 */
 
+void clean_exit P((int));
+
 #ifdef	NEWINIT
 char *gen_id = "fe";
 #endif
@@ -1394,13 +1465,14 @@ char *gen_id = "fe";
  */
 
 /* ARGSUSED */
-	void
+void
 startslave(host, autologin, autoname)
 	char *host;
 	int autologin;
 	char *autoname;
 {
 	register int i;
+	int syncpipe[2];
 #if defined(OSF) || defined(ultrix) || defined(__hpux)
 	time_t time();
 #else
@@ -1420,7 +1492,7 @@ startslave(host, autologin, autoname)
 
 	if (autologin < auth_level) {
 		fatal(net, "Authorization failed");
-		exit(1);
+		clean_exit(1);
 	}
 #ifdef TJW
 	if (autologin == 0
@@ -1429,20 +1501,43 @@ startslave(host, autologin, autoname)
 #endif
 	   ) {
 		char m[] = "Warning:  This session is not using secure authentication.\r\n";
-		write(net, m, sizeof(m));
+		WRITE(net, m, sizeof(m));
 		syslog(LOG_WARNING, "Host %s failed to negotiate strong authentication", host);
 	}
 #endif
 #endif
 
+	/*
+	 * TJW: If we have a properly authenticated username, use it
+	 * to set permissions on any files created by telnetd that
+	 * belong to the user.  Otherwise, let the "pty select hack"
+	 * handle it.
+	 */
+	if(autologin == AUTH_VALID && autoname != NULL && *autoname != '\0') {
+		struct passwd *pwd = getpwnam(autoname);
+		if(pwd != NULL) {
+			login_uid = pwd->pw_uid;
+			login_gid = pwd->pw_gid;
+			if(probe_uid_needed())
+				do_uid_callbacks(login_uid, login_gid);
+		}
+	}
+
 #ifndef	NEWINIT
 # ifdef	PARENT_DOES_UTMP
 	utmp_sig_init();
 # endif	/* PARENT_DOES_UTMP */
 
+	if(pipe(syncpipe) < 0) {
+	  fatal(net, "Error creating sychronization pipe");
+	  clean_exit(1);
+	}
+
 	if ((i = fork()) < 0)
 		fatalperror(net, "fork");
 	if (i) {
+		char c;
+
 # ifdef PARENT_DOES_UTMP
 		/*
 		 * Cray parent will create utmp entry for child and send
@@ -1452,7 +1547,11 @@ startslave(host, autologin, autoname)
 		register int pid = i;
 		void sigjob P((int));
 
+#  ifdef SETPGRP_VOID
 		setpgrp();
+#  else
+		setpgrp(0, 0);
+#  endif /* SETPGRP_VOID */
 		utmp_sig_reset();		/* reset handler to default */
 		/*
 		 * Create utmp entry for child
@@ -1479,8 +1578,24 @@ startslave(host, autologin, autoname)
 #endif
 		utmp_sig_notify(pid);
 # endif	/* PARENT_DOES_UTMP */
+
+		/* Wait for child before writing to parent side of pty. */
+		(void) close(syncpipe[1]);
+		if(read(syncpipe[0], &c, 1) <= 0) {
+		  fatal(net, "Error waiting for child initialization");
+		}
+		close(syncpipe[0]);
 	} else {
+# ifdef TLS	/* TJW: must disable TLS-based I/O in the child */
+		tls_active = 0;
+# endif /* TLS */
 		getptyslave(autologin);
+
+		/* Tell parent to continue. */
+		(void) write(syncpipe[1], "\n", 1);
+		(void) close(syncpipe[0]);
+		(void) close(syncpipe[1]);
+
 		start_login(host, autologin, autoname);
 		/*NOTREACHED*/
 	}
@@ -1492,7 +1607,7 @@ startslave(host, autologin, autoname)
 	 */
 	if ((i = open(INIT_FIFO, O_WRONLY)) < 0) {
 		char tbuf[128];
-		(void) sprintf(tbuf, "Can't open %s\n", INIT_FIFO);
+		(void) snprintf(tbuf, sizeof(tbuf), "Can't open %s\n", INIT_FIFO);
 		fatalperror(net, tbuf);
 	}
 	memset((char *)&request, 0, sizeof(request));
@@ -1515,7 +1630,7 @@ startslave(host, autologin, autoname)
 #endif /* BFTPDAEMON */
 	if (write(i, (char *)&request, sizeof(request)) < 0) {
 		char tbuf[128];
-		(void) sprintf(tbuf, "Can't write to %s\n", INIT_FIFO);
+		(void) snprintf(tbuf, sizeof(tbuf), "Can't write to %s\n", INIT_FIFO);
 		fatalperror(net, tbuf);
 	}
 	(void) close(i);
@@ -1527,8 +1642,9 @@ startslave(host, autologin, autoname)
 		if (i == 3 || n >= 0 || !gotalarm)
 			break;
 		gotalarm = 0;
-		sprintf(tbuf, "telnetd: waiting for /etc/init to start login process on %s\r\n", line);
-		(void) write(net, tbuf, strlen(tbuf));
+		snprintf(tbuf, sizeof(tbuf),
+			 "telnetd: waiting for /etc/init to start login process on %s\r\n", line);
+		(void) WRITE(net, tbuf, strlen(tbuf));
 	}
 	if (n < 0 && gotalarm)
 		fatal(net, "/etc/init didn't start login process");
@@ -1543,7 +1659,7 @@ startslave(host, autologin, autoname)
 char	*envinit[3];
 extern char **environ;
 
-	void
+void
 init_env()
 {
 	extern char *getenv();
@@ -1562,6 +1678,58 @@ init_env()
 
 #ifndef	NEWINIT
 
+# if defined(NO_LOGIN_F) && defined(LOGIN_R)
+# include <stdio.h>
+
+/* Make sure that "file" has an entry with "host" in it */
+void
+fix_hosts_equiv(file, host)
+     char *file;
+     char *host;
+{
+  FILE * fp;
+  char *p;
+  int c;
+  int makenew = 0;
+
+  if((fp = fopen(file, "r")) != NULL) {
+    p = host;
+    while((c = getc(fp)) != EOF) {
+      if(*p == '\0' && (c == ' ' || c == '\t' || c == '\r' || c == '\n'))
+	break;	/* Found it; done */
+
+      if(c == *p) {
+	++p;
+	continue;
+      }
+
+      /* Skip rest of line, reset state machine */
+      p = host;
+      while(c != '\n' && c != EOF)
+	c = getc(fp);
+      if(c == EOF)
+	break;
+    }
+
+    fclose(fp);
+
+    if(*p == '\0')
+      return;
+  }
+  else
+    makenew = 1;
+
+  if((fp = fopen(file, "a")) != NULL) {
+    fprintf(fp, "%s\n", host);
+    fclose(fp);
+    if(makenew)		/* If we created it, make it non-world-readable */
+      chmod(file, 0600);
+  }
+}
+# endif /* NO_LOGIN_F && LOGIN_R */
+
+void scrub_env();
+
 /*
  * start_login(host)
  *
@@ -1569,7 +1737,7 @@ init_env()
  * function will turn us into the login process.
  */
 
-	void
+void
 start_login(host, autologin, name)
 	char *host;
 	int autologin;
@@ -1606,8 +1774,13 @@ start_login(host, autologin, name)
 	utmpx.ut_id[3] = SC_WILDC;
 	utmpx.ut_type = LOGIN_PROCESS;
 	(void) time(&utmpx.ut_tv.tv_sec);
+# ifdef HAVE_MAKEUTX
 	if (makeutx(&utmpx) == NULL)
 		fatal(net, "makeutx failed");
+# else
+	if (pututxline(&utmpx) == NULL)
+		fatal(net, "pututxline failed");
+# endif /* HAVE_MAKEUTX */
 #endif
 
 	scrub_env();
@@ -1684,10 +1857,14 @@ start_login(host, autologin, name)
 	if (require_SecurID)
 		argv = addarg(argv, "-s");
 #endif
-#if	defined (AUTHENTICATION)
-	if (auth_level >= 0 && autologin == AUTH_VALID) {
+#if	defined (AUTHENTICATION) || defined(TLS)
+	if (auth_level >= 0 && autologin == AUTH_VALID && name[0]) {
 # if	!defined(NO_LOGIN_F)
+#  ifdef LOGIN_CAP_F
+		argv = addarg(argv, "-F");
+#  else
 		argv = addarg(argv, "-f");
+#  endif /* LOGIN_CAP_F */
 		argv = addarg(argv, name);
 # else
 #  if defined(LOGIN_R)
@@ -1701,7 +1878,7 @@ start_login(host, autologin, name)
 
 		if (pty > 2) {
 			register char *cp;
-			char speed[128];
+			char speed[1024];
 			int isecho, israw, xpty, len;
 			extern int def_rspeed;
 #  ifndef LOGIN_HOST
@@ -1718,6 +1895,14 @@ start_login(host, autologin, name)
 			 */
 #			define LOGIN_HOST "localhost"
 #  endif
+#  ifndef PATH_HEQUIV
+			/*
+			 * Make sure our LOGIN_HOST is accepted as a
+			 * trusted host.
+			 */
+#			define PATH_HEQUIV "/etc/hosts.equiv"
+#  endif
+			fix_hosts_equiv(PATH_HEQUIV, LOGIN_HOST);
 			argv = addarg(argv, "-r");
 			argv = addarg(argv, LOGIN_HOST);
 
@@ -1738,7 +1923,8 @@ start_login(host, autologin, name)
 			len = strlen(name)+1;
 			write(xpty, name, len);
 			write(xpty, name, len);
-			sprintf(speed, "%s/%d", (cp = getenv("TERM")) ? cp : "",
+			snprintf(speed, sizeof(speed),
+				 "%s/%d", (cp = getenv("TERM")) ? cp : "",
 				(def_rspeed > 0) ? def_rspeed : 9600);
 			len = strlen(speed)+1;
 			write(xpty, speed, len);
@@ -1748,6 +1934,7 @@ start_login(host, autologin, name)
 				tty_setecho(isecho);
 				tty_setraw(israw);
 				set_termbuf();
+#ifndef SOLARIS
 				if (!israw) {
 					/*
 					 * Write a newline to ensure
@@ -1756,6 +1943,7 @@ start_login(host, autologin, name)
 					 */
 					write(xpty, "\n", 1);
 				}
+#endif /* SOLARIS */
 			}
 			pty = xpty;
 		}
@@ -1767,13 +1955,6 @@ start_login(host, autologin, name)
 #endif
 	if (getenv("USER")) {
 		argv = addarg(argv, getenv("USER"));
-#if	defined(LOGIN_ARGS) && defined(NO_LOGIN_P)
-		{
-			register char **cpp;
-			for (cpp = environ; *cpp; cpp++)
-				argv = addarg(argv, *cpp);
-		}
-#endif
 		/*
 		 * Assume that login will set the USER variable
 		 * correctly.  For SysV systems, this means that
@@ -1796,25 +1977,26 @@ start_login(host, autologin, name)
 		}
 	}
 #endif	/* SOLARIS */
+#if	defined(LOGIN_ARGS) && defined(NO_LOGIN_P)
+	{
+		register char **cpp;
+		for (cpp = environ; *cpp; cpp++)
+			argv = addarg(argv, *cpp);
+	}
+#endif
 #if	defined(AUTHENTICATION) && defined(NO_LOGIN_F) && defined(LOGIN_R)
 	if (pty > 2)
 		close(pty);
 #endif
 	closelog();
-	/*
-	 * This sleep(1) is in here so that telnetd can
-	 * finish up with the tty.  There's a race condition
-	 * the login banner message gets lost...
-	 */
-	sleep(1);
-	execv(_PATH_LOGIN, argv);
+	execv(login_program, argv);
 
-	syslog(LOG_ERR, "%s: %m\n", _PATH_LOGIN);
-	fatalperror(net, _PATH_LOGIN);
+	syslog(LOG_ERR, "%s: %m", login_program);
+	fatalperror(net, login_program);
 	/*NOTREACHED*/
 }
 
-	char **
+char **
 addarg(argv, val)
 	register char **argv;
 	register char *val;
@@ -1854,6 +2036,7 @@ addarg(argv, val)
  * Remove a few things from the environment that
  * don't need to be there.
  */
+void
 scrub_env()
 {
 	register char **cpp, **cpp2;
@@ -1875,7 +2058,7 @@ scrub_env()
  * clean up anything that needs to be cleaned up.
  */
 	/* ARGSUSED */
-	void
+void
 cleanup(sig)
 	int sig;
 {
@@ -1891,18 +2074,45 @@ cleanup(sig)
 	*p = 'p';
 	(void)chmod(line, 0666);
 	(void)chown(line, 0, 0);
+#ifdef TLS
+	tls_shutdown();
+	tls_cleanup();
+#endif
+#ifdef FWD_X
+	fwdx_cleanup();
+#endif
 	(void) shutdown(net, 2);
+	if (netobuf)
+	    free(netobuf);
 	exit(1);
 # else
 	void rmut();
 	rmut();
+#  ifdef HAVE_VHANGUP
 	vhangup();	/* XXX */
+#  endif /* HAVE_VHANGUP */
+#ifdef TLS
+	tls_shutdown();
+	tls_cleanup();
+#endif
+#ifdef FWD_X
+	fwdx_cleanup();
+#endif
 	(void) shutdown(net, 2);
 	exit(1);
 # endif
 #else	/* PARENT_DOES_UTMP */
 # ifdef	NEWINIT
+#ifdef TLS
+	tls_shutdown();
+	tls_cleanup();
+#endif
+#ifdef FWD_X
+	fwdx_cleanup();
+#endif
 	(void) shutdown(net, 2);
+	if (netobuf)
+	    free(netobuf);
 	exit(1);
 # else	/* NEWINIT */
 #  ifdef CRAY
@@ -1955,16 +2165,40 @@ cleanup(sig)
 #  endif /* CRAY */
 	rmut(line);
 	close(pty);
+#ifdef TLS
+	tls_shutdown();
+	tls_cleanup();
+#endif
+#ifdef FWD_X
+	fwdx_cleanup();
+#endif
 	(void) shutdown(net, 2);
 #  ifdef CRAY
 	if (t == 0)
 		cleantmp(&wtmp);
 #  endif /* CRAY */
+	if (netobuf)
+	    free(netobuf);
 	exit(1);
 # endif	/* NEWINT */
 #endif	/* PARENT_DOES_UTMP */
 }
 
+void clean_exit(err)
+	int err;
+{
+#ifdef TLS
+	tls_shutdown();
+	tls_cleanup();
+#endif
+#ifdef FWD_X
+	fwdx_cleanup();
+#endif
+	if (netobuf)
+	    free(netobuf);
+	exit(err);
+}
+
 #if defined(PARENT_DOES_UTMP) && !defined(NEWINIT)
 /*
  * _utmp_sig_rcv
@@ -1979,7 +2213,7 @@ cleanup(sig)
 static int caught=0;		/* NZ when signal intercepted */
 static void (*func)();		/* address of previous handler */
 
-	void
+void
 _utmp_sig_rcv(sig)
 	int sig;
 {
@@ -1987,7 +2221,7 @@ _utmp_sig_rcv(sig)
 	(void) signal(SIGUSR1, func);
 }
 
-	void
+void
 utmp_sig_init()
 {
 	/*
@@ -1997,7 +2231,7 @@ utmp_sig_init()
 		fatalperror(net, "telnetd/signal");
 }
 
-	void
+void
 utmp_sig_reset()
 {
 	(void) signal(SIGUSR1, func);	/* reset handler to default */
@@ -2008,7 +2242,7 @@ utmp_sig_reset()
 # define sigon() /* do nothing */
 # endif
 
-	void
+void
 utmp_sig_wait()
 {
 	/*
@@ -2026,8 +2260,9 @@ utmp_sig_wait()
 #endif
 }
 
-	void
+void
 utmp_sig_notify(pid)
+	int pid;
 {
 	kill(pid, SIGUSR1);
 }
@@ -2036,7 +2271,7 @@ utmp_sig_notify(pid)
 static int gotsigjob = 0;
 
 	/*ARGSUSED*/
-	void
+void
 sigjob(sig)
 	int sig;
 {
@@ -2058,7 +2293,7 @@ sigjob(sig)
  *		to find the correct $TMPDIR to cleanup.
  */
 
-	struct utmp *
+struct utmp *
 jid_getutid(jid)
 	int jid;
 {
@@ -2082,7 +2317,7 @@ jid_getutid(jid)
  * when this is called the second time it will wait
  * for the signal that the job is done.
  */
-	int
+int
 cleantmp(wtp)
 	register struct utmp *wtp;
 {
@@ -2123,7 +2358,7 @@ cleantmp(wtp)
 	return(ret);
 }
 
-	int
+int
 jobend(jid, path, user)
 	register int jid;
 	register char *path;
@@ -2186,6 +2421,7 @@ jobend(jid, path, user)
 /*
  * Fork a child process to clean up the TMPDIR
  */
+int
 cleantmpdir(jid, tpath, user)
 	register int jid;
 	register char *tpath;
@@ -2193,12 +2429,12 @@ cleantmpdir(jid, tpath, user)
 {
 	switch(fork()) {
 	case -1:
-		syslog(LOG_ERR, "TMPDIR cleanup(%s): fork() failed: %m\n",
+		syslog(LOG_ERR, "TMPDIR cleanup(%s): fork() failed: %m",
 							tpath);
 		break;
 	case 0:
 		execl(CLEANTMPCMD, CLEANTMPCMD, user, tpath, 0);
-		syslog(LOG_ERR, "TMPDIR cleanup(%s): execl(%s) failed: %m\n",
+		syslog(LOG_ERR, "TMPDIR cleanup(%s): execl(%s) failed: %m",
 							tpath, CLEANTMPCMD);
 		exit(1);
 	default:
@@ -2221,7 +2457,7 @@ cleantmpdir(jid, tpath, user)
 
 #ifdef	UTMPX
 #ifndef __hpux
-	void
+void
 rmut()
 {
 	register f;
@@ -2240,11 +2476,32 @@ rmut()
 	utxp = getutxline(&utmpx);
 	if (utxp) {
 		utxp->ut_type = DEAD_PROCESS;
+#ifdef __USE_GNU
 		utxp->ut_exit.e_termination = 0;
 		utxp->ut_exit.e_exit = 0;
+#endif /* __USE_GNU */
 		(void) time(&utxp->ut_tv.tv_sec);
 		utxp->ut_tv.tv_usec = 0;
+#ifdef HAVE_MODUTX
 		modutx(utxp);
+#elif defined(HAVE_UPDWTMPX)
+		pututxline(utxp);
+		updwtmpx(WTMPX_FILE, utxp);
+#else
+		f = open(wtmpf, O_WRONLY|O_APPEND);
+		if (f >= 0) {
+			SCPYN(wtmp.ut_line, line+5);
+			SCPYN(wtmp.UT_USER, "");
+			SCPYN(wtmp.ut_host, "");
+#if (!defined(NO_WTMPX) && defined(HAVE_UTMPX_UT_TV)) || (defined(NO_WTMPX) && defined(HAVE_UTMP_UT_TV))
+			(void) time(&wtmp.ut_tv.tv_sec);
+#else
+			(void) time(&wtmp.ut_time);
+#endif
+			(void) write(f, (char *)&wtmp, sizeof(wtmp));
+			(void) close(f);
+		}
+#endif
 	}
 	endutxent();
 }  /* end of rmut */
@@ -2252,7 +2509,7 @@ rmut()
 #endif
 
 #if	!defined(UTMPX) && !(defined(CRAY) || defined(__hpux)) && BSD <= 43
-	void
+void
 rmut()
 {
 	register f;
@@ -2278,7 +2535,11 @@ rmut()
 				(void) lseek(f, ((long)u)-((long)utmp), L_SET);
 				SCPYN(u->ut_name, "");
 				SCPYN(u->ut_host, "");
+#ifdef HAVE_UTMP_UT_TV
+				(void) time(&u->ut_tv.tv_sec);
+#else
 				(void) time(&u->ut_time);
+#endif
 				(void) write(f, (char *)u, sizeof(wtmp));
 				found++;
 			}
@@ -2289,7 +2550,7 @@ rmut()
 		f = open(wtmpf, O_WRONLY|O_APPEND);
 		if (f >= 0) {
 			SCPYN(wtmp.ut_line, line+5);
-			SCPYN(wtmp.ut_name, "");
+			SCPYN(wtmp.UT_USER, "");
 			SCPYN(wtmp.ut_host, "");
 			(void) time(&wtmp.ut_time);
 			(void) write(f, (char *)&wtmp, sizeof(wtmp));
@@ -2305,9 +2566,9 @@ rmut()
 #endif	/* CRAY */
 
 #ifdef __hpux
-
+int
 rmut (line)
-char *line;
+	char *line;
 {
 	struct utmp utmp;
 	struct utmp *utptr;
