@@ -35,6 +35,20 @@
 static char sccsid[] = "@(#)utility.c	8.4 (Berkeley) 5/30/95";
 #endif /* not lint */
 
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
+
+#ifdef TLS
+#define SEND  tls_send
+#define WRITE tls_write
+#define READ  tls_read
+#else
+#define SEND  send
+#define WRITE write
+#define READ  read
+#endif
+
 #define PRINTOPTIONS
 #include "telnetd.h"
 
@@ -56,20 +70,20 @@ ttloop()
 {
     void netflush();
 
-    DIAG(TD_REPORT, {sprintf(nfrontp, "td: ttloop\r\n");
+    DIAG(TD_REPORT, {net_printf("td: ttloop\r\n");
 		     nfrontp += strlen(nfrontp);});
     if (nfrontp-nbackp) {
 	netflush();
     }
-    ncc = read(net, netibuf, sizeof netibuf);
+    ncc = READ(net, netibuf, sizeof netibuf);
     if (ncc < 0) {
-	syslog(LOG_INFO, "ttloop:  read: %m\n");
-	exit(1);
+	syslog(LOG_INFO, "ttloop:  read: %m");
+	clean_exit(1);
     } else if (ncc == 0) {
-	syslog(LOG_INFO, "ttloop:  peer died: %m\n");
-	exit(1);
+	syslog(LOG_INFO, "ttloop:  peer died: %m");
+	clean_exit(1);
     }
-    DIAG(TD_REPORT, {sprintf(nfrontp, "td: ttloop read %d chars\r\n", ncc);
+    DIAG(TD_REPORT, {net_printf("td: ttloop read %d chars\r\n", ncc);
 		     nfrontp += strlen(nfrontp);});
     netip = netibuf;
     telrcv();			/* state machine */
@@ -113,7 +127,7 @@ ptyflush()
 
 	if ((n = pfrontp - pbackp) > 0) {
 		DIAG((TD_REPORT | TD_PTYDATA),
-			{ sprintf(nfrontp, "td: ptyflush %d chars\r\n", n);
+			{ net_printf("td: ptyflush %d chars\r\n", n);
 			  nfrontp += strlen(nfrontp); });
 		DIAG(TD_PTYDATA, printdata("pd", pbackp, n));
 		n = write(pty, pbackp, n);
@@ -123,7 +137,8 @@ ptyflush()
 			return;
 		cleanup(0);
 	}
-	pbackp += n;
+	else
+		pbackp += n;
 	if (pbackp == pfrontp)
 		pbackp = pfrontp = ptyobuf;
 }
@@ -246,7 +261,7 @@ netflush()
 
     if ((n = nfrontp - nbackp) > 0) {
 	DIAG(TD_REPORT,
-	    { sprintf(nfrontp, "td: netflush %d chars\r\n", n);
+	    { net_printf("td: netflush %d chars\r\n", n);
 	      n += strlen(nfrontp);  /* get count first */
 	      nfrontp += strlen(nfrontp);  /* then move pointer */
 	    });
@@ -265,7 +280,7 @@ netflush()
 	 * write the entire buffer in non-OOB mode.
 	 */
 	if ((neturg == 0) || (not42 == 0)) {
-	    n = write(net, nbackp, n);	/* normal write */
+	    n = WRITE(net, nbackp, n);	/* normal write */
 	} else {
 	    n = neturg - nbackp;
 	    /*
@@ -277,9 +292,9 @@ netflush()
 	     * rather than the Unix philosophy of OOB data).
 	     */
 	    if (n > 1) {
-		n = send(net, nbackp, n-1, 0);	/* send URGENT all by itself */
+		n = SEND(net, nbackp, n-1, 0);	/* send URGENT all by itself */
 	    } else {
-		n = send(net, nbackp, n, MSG_OOB);	/* URGENT data */
+		n = SEND(net, nbackp, n, MSG_OOB);	/* URGENT data */
 	    }
 	}
     }
@@ -322,7 +337,7 @@ writenet(ptr, len)
 	register int len;
 {
 	/* flush buffer if no room for new data) */
-	if ((&netobuf[BUFSIZ] - nfrontp) < len) {
+	if ((&netobuf[netobuf_size] - nfrontp) < len) {
 		/* if this fails, don't worry, buffer is a little big */
 		netflush();
 	}
@@ -345,7 +360,7 @@ fatal(f, msg)
 {
 	char buf[BUFSIZ];
 
-	(void) sprintf(buf, "telnetd: %s.\r\n", msg);
+	(void) snprintf(buf, sizeof(buf), "telnetd: %s.\r\n", msg);
 #ifdef	ENCRYPTION
 	if (encrypt_output) {
 		/*
@@ -356,9 +371,9 @@ fatal(f, msg)
 		netflush();
 	}
 #endif	/* ENCRYPTION */
-	(void) write(f, buf, (int)strlen(buf));
+	(void) WRITE(f, buf, (int)strlen(buf));
 	sleep(1);	/*XXX*/
-	exit(1);
+	clean_exit(1);
 }
 
 	void
@@ -368,7 +383,7 @@ fatalperror(f, msg)
 {
 	char buf[BUFSIZ], *strerror();
 
-	(void) sprintf(buf, "%s: %s", msg, strerror(errno));
+	(void) snprintf(buf, sizeof(buf), "%s: %s", msg, strerror(errno));
 	fatal(f, buf);
 }
 
@@ -380,7 +395,9 @@ edithost(pat, host)
 	register char *host;
 {
 	register char *res = editedhost;
-	/*char *strncpy();*/
+#ifndef strncpy
+	char *strncpy();
+#endif
 
 	if (!pat)
 		pat = "";
@@ -416,12 +433,12 @@ edithost(pat, host)
 }
 
 static char *putlocation;
+static char lastch;
 
 	void
 putstr(s)
 	register char *s;
 {
-
 	while (*s)
 		putchr(*s++);
 }
@@ -430,7 +447,9 @@ putstr(s)
 putchr(cc)
 	int cc;
 {
-	*putlocation++ = cc;
+	if(cc == '\n' && lastch != '\r')
+		*putlocation++ = '\r';
+	lastch = *putlocation++ = cc;
 }
 
 /*
@@ -448,15 +467,23 @@ putf(cp, where)
 	char *slash;
 	time_t t;
 	char db[100];
+#ifdef HAVE_UNAME
+        struct utsname u;
+#endif /* HAVE_UNAME */
 #ifdef	STREAMSPTY
+#ifndef strchr
 	extern char *strchr();
+#endif
 #else
+#ifndef strrchr
 	extern char *strrchr();
+#endif
 #endif
 
 	putlocation = where;
+	lastch = '\0';
 
-	while (*cp) {
+	while(*cp) {
 		if (*cp != '%') {
 			putchr(*cp++);
 			continue;
@@ -486,6 +513,28 @@ putf(cp, where)
 			putstr(db);
 			break;
 
+#ifdef HAVE_UNAME
+		case 's':
+			if (uname(&u) == 0)
+			    putstr(u.sysname);
+		        break;
+		case 'n':
+			if (uname(&u) == 0)
+			    putstr(u.nodename);
+		        break;
+		case 'r':
+			if (uname(&u) == 0)
+			    putstr(u.release);
+		        break;
+		case 'v':
+			if (uname(&u) == 0)
+			    putstr(u.version);
+		        break;
+		case 'm':
+			if (uname(&u) == 0)
+			    putstr(u.machine);
+		        break;
+#endif /* HAVE_UNAME */
 		case '%':
 			putchr('%');
 			break;
@@ -504,11 +553,11 @@ printoption(fmt, option)
 	register int option;
 {
 	if (TELOPT_OK(option))
-		sprintf(nfrontp, "%s %s\r\n", fmt, TELOPT(option));
+		net_printf("%s %s\r\n", fmt, TELOPT(option));
 	else if (TELCMD_OK(option))
-		sprintf(nfrontp, "%s %s\r\n", fmt, TELCMD(option));
+		net_printf("%s %s\r\n", fmt, TELCMD(option));
 	else
-		sprintf(nfrontp, "%s %d\r\n", fmt, option);
+		net_printf("%s %d\r\n", fmt, option);
 	nfrontp += strlen(nfrontp);
 	return;
 }
@@ -526,7 +575,7 @@ printsub(direction, pointer, length)
 		return;
 
 	if (direction) {
-	    sprintf(nfrontp, "td: %s suboption ",
+	    net_printf("td: %s suboption ",
 					direction == '<' ? "recv" : "send");
 	    nfrontp += strlen(nfrontp);
 	    if (length >= 3) {
@@ -536,72 +585,71 @@ printsub(direction, pointer, length)
 		j = pointer[length-1];
 
 		if (i != IAC || j != SE) {
-		    sprintf(nfrontp, "(terminated by ");
+		    net_printf("(terminated by ");
 		    nfrontp += strlen(nfrontp);
 		    if (TELOPT_OK(i))
-			sprintf(nfrontp, "%s ", TELOPT(i));
+			net_printf("%s ", TELOPT(i));
 		    else if (TELCMD_OK(i))
-			sprintf(nfrontp, "%s ", TELCMD(i));
+			net_printf("%s ", TELCMD(i));
 		    else
-			sprintf(nfrontp, "%d ", i);
+			net_printf("%d ", i);
 		    nfrontp += strlen(nfrontp);
 		    if (TELOPT_OK(j))
-			sprintf(nfrontp, "%s", TELOPT(j));
+			net_printf("%s", TELOPT(j));
 		    else if (TELCMD_OK(j))
-			sprintf(nfrontp, "%s", TELCMD(j));
+			net_printf("%s", TELCMD(j));
 		    else
-			sprintf(nfrontp, "%d", j);
+			net_printf("%d", j);
 		    nfrontp += strlen(nfrontp);
-		    sprintf(nfrontp, ", not IAC SE!) ");
+		    net_printf(", not IAC SE!) ");
 		    nfrontp += strlen(nfrontp);
 		}
 	    }
 	    length -= 2;
 	}
 	if (length < 1) {
-	    sprintf(nfrontp, "(Empty suboption??\?)");
+	    net_printf("(Empty suboption??\?)");
 	    nfrontp += strlen(nfrontp);
 	    return;
 	}
 	switch (pointer[0]) {
 	case TELOPT_TTYPE:
-	    sprintf(nfrontp, "TERMINAL-TYPE ");
+	    net_printf("TERMINAL-TYPE ");
 	    nfrontp += strlen(nfrontp);
 	    switch (pointer[1]) {
 	    case TELQUAL_IS:
-		sprintf(nfrontp, "IS \"%.*s\"", length-2, (char *)pointer+2);
+		net_printf("IS \"%.*s\"", length-2, (char *)pointer+2);
 		break;
 	    case TELQUAL_SEND:
-		sprintf(nfrontp, "SEND");
+		net_printf("SEND");
 		break;
 	    default:
-		sprintf(nfrontp,
-				"- unknown qualifier %d (0x%x).",
+		net_printf("- unknown qualifier %d (0x%x).",
 				pointer[1], pointer[1]);
 	    }
 	    nfrontp += strlen(nfrontp);
 	    break;
 	case TELOPT_TSPEED:
-	    sprintf(nfrontp, "TERMINAL-SPEED");
+	    net_printf("TERMINAL-SPEED");
 	    nfrontp += strlen(nfrontp);
 	    if (length < 2) {
-		sprintf(nfrontp, " (empty suboption??\?)");
+		net_printf(" (empty suboption??\?)");
 		nfrontp += strlen(nfrontp);
 		break;
 	    }
 	    switch (pointer[1]) {
 	    case TELQUAL_IS:
-		sprintf(nfrontp, " IS %.*s", length-2, (char *)pointer+2);
+		net_printf(" IS %.*s", length-2, (char *)pointer+2);
 		nfrontp += strlen(nfrontp);
 		break;
 	    default:
 		if (pointer[1] == 1)
-		    sprintf(nfrontp, " SEND");
+		    net_printf(" SEND");
 		else
-		    sprintf(nfrontp, " %d (unknown)", pointer[1]);
+		    net_printf(" %d (unknown)", pointer[1]);
 		nfrontp += strlen(nfrontp);
 		for (i = 2; i < length; i++) {
-		    sprintf(nfrontp, " ?%d?", pointer[i]);
+		    net_printf(" ?%d?", pointer[i]);
 		    nfrontp += strlen(nfrontp);
 		}
 		break;
@@ -609,105 +657,105 @@ printsub(direction, pointer, length)
 	    break;
 
 	case TELOPT_LFLOW:
-	    sprintf(nfrontp, "TOGGLE-FLOW-CONTROL");
+	    net_printf("TOGGLE-FLOW-CONTROL");
 	    nfrontp += strlen(nfrontp);
 	    if (length < 2) {
-		sprintf(nfrontp, " (empty suboption??\?)");
+		net_printf(" (empty suboption??\?)");
 		nfrontp += strlen(nfrontp);
 		break;
 	    }
 	    switch (pointer[1]) {
 	    case LFLOW_OFF:
-		sprintf(nfrontp, " OFF"); break;
+		net_printf(" OFF"); break;
 	    case LFLOW_ON:
-		sprintf(nfrontp, " ON"); break;
+		net_printf(" ON"); break;
 	    case LFLOW_RESTART_ANY:
-		sprintf(nfrontp, " RESTART-ANY"); break;
+		net_printf(" RESTART-ANY"); break;
 	    case LFLOW_RESTART_XON:
-		sprintf(nfrontp, " RESTART-XON"); break;
+		net_printf(" RESTART-XON"); break;
 	    default:
-		sprintf(nfrontp, " %d (unknown)", pointer[1]);
+		net_printf(" %d (unknown)", pointer[1]);
 	    }
 	    nfrontp += strlen(nfrontp);
 	    for (i = 2; i < length; i++) {
-		sprintf(nfrontp, " ?%d?", pointer[i]);
+		net_printf(" ?%d?", pointer[i]);
 		nfrontp += strlen(nfrontp);
 	    }
 	    break;
 
 	case TELOPT_NAWS:
-	    sprintf(nfrontp, "NAWS");
+	    net_printf("NAWS");
 	    nfrontp += strlen(nfrontp);
 	    if (length < 2) {
-		sprintf(nfrontp, " (empty suboption??\?)");
+		net_printf(" (empty suboption??\?)");
 		nfrontp += strlen(nfrontp);
 		break;
 	    }
 	    if (length == 2) {
-		sprintf(nfrontp, " ?%d?", pointer[1]);
+		net_printf(" ?%d?", pointer[1]);
 		nfrontp += strlen(nfrontp);
 		break;
 	    }
-	    sprintf(nfrontp, " %d %d (%d)",
+	    net_printf(" %d %d (%d)",
 		pointer[1], pointer[2],
 		(int)((((unsigned int)pointer[1])<<8)|((unsigned int)pointer[2])));
 	    nfrontp += strlen(nfrontp);
 	    if (length == 4) {
-		sprintf(nfrontp, " ?%d?", pointer[3]);
+		net_printf(" ?%d?", pointer[3]);
 		nfrontp += strlen(nfrontp);
 		break;
 	    }
-	    sprintf(nfrontp, " %d %d (%d)",
+	    net_printf(" %d %d (%d)",
 		pointer[3], pointer[4],
 		(int)((((unsigned int)pointer[3])<<8)|((unsigned int)pointer[4])));
 	    nfrontp += strlen(nfrontp);
 	    for (i = 5; i < length; i++) {
-		sprintf(nfrontp, " ?%d?", pointer[i]);
+		net_printf(" ?%d?", pointer[i]);
 		nfrontp += strlen(nfrontp);
 	    }
 	    break;
 
 	case TELOPT_LINEMODE:
-	    sprintf(nfrontp, "LINEMODE ");
+	    net_printf("LINEMODE ");
 	    nfrontp += strlen(nfrontp);
 	    if (length < 2) {
-		sprintf(nfrontp, " (empty suboption??\?)");
+		net_printf(" (empty suboption??\?)");
 		nfrontp += strlen(nfrontp);
 		break;
 	    }
 	    switch (pointer[1]) {
 	    case WILL:
-		sprintf(nfrontp, "WILL ");
+		net_printf("WILL ");
 		goto common;
 	    case WONT:
-		sprintf(nfrontp, "WONT ");
+		net_printf("WONT ");
 		goto common;
 	    case DO:
-		sprintf(nfrontp, "DO ");
+		net_printf("DO ");
 		goto common;
 	    case DONT:
-		sprintf(nfrontp, "DONT ");
+		net_printf("DONT ");
 	    common:
 		nfrontp += strlen(nfrontp);
 		if (length < 3) {
-		    sprintf(nfrontp, "(no option??\?)");
+		    net_printf("(no option??\?)");
 		    nfrontp += strlen(nfrontp);
 		    break;
 		}
 		switch (pointer[2]) {
 		case LM_FORWARDMASK:
-		    sprintf(nfrontp, "Forward Mask");
+		    net_printf("Forward Mask");
 		    nfrontp += strlen(nfrontp);
 		    for (i = 3; i < length; i++) {
-			sprintf(nfrontp, " %x", pointer[i]);
+			net_printf(" %x", pointer[i]);
 			nfrontp += strlen(nfrontp);
 		    }
 		    break;
 		default:
-		    sprintf(nfrontp, "%d (unknown)", pointer[2]);
+		    net_printf("%d (unknown)", pointer[2]);
 		    nfrontp += strlen(nfrontp);
 		    for (i = 3; i < length; i++) {
-			sprintf(nfrontp, " %d", pointer[i]);
+			net_printf(" %d", pointer[i]);
 			nfrontp += strlen(nfrontp);
 		    }
 		    break;
@@ -715,80 +763,80 @@ printsub(direction, pointer, length)
 		break;
 
 	    case LM_SLC:
-		sprintf(nfrontp, "SLC");
+		net_printf("SLC");
 		nfrontp += strlen(nfrontp);
 		for (i = 2; i < length - 2; i += 3) {
 		    if (SLC_NAME_OK(pointer[i+SLC_FUNC]))
-			sprintf(nfrontp, " %s", SLC_NAME(pointer[i+SLC_FUNC]));
+			net_printf(" %s", SLC_NAME(pointer[i+SLC_FUNC]));
 		    else
-			sprintf(nfrontp, " %d", pointer[i+SLC_FUNC]);
+			net_printf(" %d", pointer[i+SLC_FUNC]);
 		    nfrontp += strlen(nfrontp);
 		    switch (pointer[i+SLC_FLAGS]&SLC_LEVELBITS) {
 		    case SLC_NOSUPPORT:
-			sprintf(nfrontp, " NOSUPPORT"); break;
+			net_printf(" NOSUPPORT"); break;
 		    case SLC_CANTCHANGE:
-			sprintf(nfrontp, " CANTCHANGE"); break;
+			net_printf(" CANTCHANGE"); break;
 		    case SLC_VARIABLE:
-			sprintf(nfrontp, " VARIABLE"); break;
+			net_printf(" VARIABLE"); break;
 		    case SLC_DEFAULT:
-			sprintf(nfrontp, " DEFAULT"); break;
+			net_printf(" DEFAULT"); break;
 		    }
 		    nfrontp += strlen(nfrontp);
-		    sprintf(nfrontp, "%s%s%s",
+		    net_printf("%s%s%s",
 			pointer[i+SLC_FLAGS]&SLC_ACK ? "|ACK" : "",
 			pointer[i+SLC_FLAGS]&SLC_FLUSHIN ? "|FLUSHIN" : "",
 			pointer[i+SLC_FLAGS]&SLC_FLUSHOUT ? "|FLUSHOUT" : "");
 		    nfrontp += strlen(nfrontp);
 		    if (pointer[i+SLC_FLAGS]& ~(SLC_ACK|SLC_FLUSHIN|
 						SLC_FLUSHOUT| SLC_LEVELBITS)) {
-			sprintf(nfrontp, "(0x%x)", pointer[i+SLC_FLAGS]);
+			net_printf("(0x%x)", pointer[i+SLC_FLAGS]);
 			nfrontp += strlen(nfrontp);
 		    }
-		    sprintf(nfrontp, " %d;", pointer[i+SLC_VALUE]);
+		    net_printf(" %d;", pointer[i+SLC_VALUE]);
 		    nfrontp += strlen(nfrontp);
 		    if ((pointer[i+SLC_VALUE] == IAC) &&
 			(pointer[i+SLC_VALUE+1] == IAC))
 				i++;
 		}
 		for (; i < length; i++) {
-		    sprintf(nfrontp, " ?%d?", pointer[i]);
+		    net_printf(" ?%d?", pointer[i]);
 		    nfrontp += strlen(nfrontp);
 		}
 		break;
 
 	    case LM_MODE:
-		sprintf(nfrontp, "MODE ");
+		net_printf("MODE ");
 		nfrontp += strlen(nfrontp);
 		if (length < 3) {
-		    sprintf(nfrontp, "(no mode??\?)");
+		    net_printf("(no mode??\?)");
 		    nfrontp += strlen(nfrontp);
 		    break;
 		}
 		{
 		    char tbuf[32];
-		    sprintf(tbuf, "%s%s%s%s%s",
+		    snprintf(tbuf, sizeof(tbuf), "%s%s%s%s%s",
 			pointer[2]&MODE_EDIT ? "|EDIT" : "",
 			pointer[2]&MODE_TRAPSIG ? "|TRAPSIG" : "",
 			pointer[2]&MODE_SOFT_TAB ? "|SOFT_TAB" : "",
 			pointer[2]&MODE_LIT_ECHO ? "|LIT_ECHO" : "",
 			pointer[2]&MODE_ACK ? "|ACK" : "");
-		    sprintf(nfrontp, "%s", tbuf[1] ? &tbuf[1] : "0");
+		    net_printf("%s", tbuf[1] ? &tbuf[1] : "0");
 		    nfrontp += strlen(nfrontp);
 		}
 		if (pointer[2]&~(MODE_EDIT|MODE_TRAPSIG|MODE_ACK)) {
-		    sprintf(nfrontp, " (0x%x)", pointer[2]);
+		    net_printf(" (0x%x)", pointer[2]);
 		    nfrontp += strlen(nfrontp);
 		}
 		for (i = 3; i < length; i++) {
-		    sprintf(nfrontp, " ?0x%x?", pointer[i]);
+		    net_printf(" ?0x%x?", pointer[i]);
 		    nfrontp += strlen(nfrontp);
 		}
 		break;
 	    default:
-		sprintf(nfrontp, "%d (unknown)", pointer[1]);
+		net_printf("%d (unknown)", pointer[1]);
 		nfrontp += strlen(nfrontp);
 		for (i = 2; i < length; i++) {
-		    sprintf(nfrontp, " %d", pointer[i]);
+		    net_printf(" %d", pointer[i]);
 		    nfrontp += strlen(nfrontp);
 		}
 	    }
@@ -798,23 +846,23 @@ printsub(direction, pointer, length)
 	    register char *cp;
 	    register int j, k;
 
-	    sprintf(nfrontp, "STATUS");
+	    net_printf("STATUS");
 	    nfrontp += strlen(nfrontp);
 
 	    switch (pointer[1]) {
 	    default:
 		if (pointer[1] == TELQUAL_SEND)
-		    sprintf(nfrontp, " SEND");
+		    net_printf(" SEND");
 		else
-		    sprintf(nfrontp, " %d (unknown)", pointer[1]);
+		    net_printf(" %d (unknown)", pointer[1]);
 		nfrontp += strlen(nfrontp);
 		for (i = 2; i < length; i++) {
-		    sprintf(nfrontp, " ?%d?", pointer[i]);
+		    net_printf(" ?%d?", pointer[i]);
 		    nfrontp += strlen(nfrontp);
 		}
 		break;
 	    case TELQUAL_IS:
-		sprintf(nfrontp, " IS\r\n");
+		net_printf(" IS\r\n");
 		nfrontp += strlen(nfrontp);
 
 		for (i = 2; i < length; i++) {
@@ -826,17 +874,17 @@ printsub(direction, pointer, length)
 		    common2:
 			i++;
 			if (TELOPT_OK(pointer[i]))
-			    sprintf(nfrontp, " %s %s", cp, TELOPT(pointer[i]));
+			    net_printf(" %s %s", cp, TELOPT(pointer[i]));
 			else
-			    sprintf(nfrontp, " %s %d", cp, pointer[i]);
+			    net_printf(" %s %d", cp, pointer[i]);
 			nfrontp += strlen(nfrontp);
 
-			sprintf(nfrontp, "\r\n");
+			net_printf("\r\n");
 			nfrontp += strlen(nfrontp);
 			break;
 
 		    case SB:
-			sprintf(nfrontp, " SB ");
+			net_printf(" SB ");
 			nfrontp += strlen(nfrontp);
 			i++;
 			j = k = i;
@@ -853,19 +901,19 @@ printsub(direction, pointer, length)
 			}
 			printsub(0, &pointer[i], k - i);
 			if (i < length) {
-			    sprintf(nfrontp, " SE");
+			    net_printf(" SE");
 			    nfrontp += strlen(nfrontp);
 			    i = j;
 			} else
 			    i = j - 1;
 
-			sprintf(nfrontp, "\r\n");
+			net_printf("\r\n");
 			nfrontp += strlen(nfrontp);
 
 			break;
 
 		    default:
-			sprintf(nfrontp, " %d", pointer[i]);
+			net_printf(" %d", pointer[i]);
 			nfrontp += strlen(nfrontp);
 			break;
 		    }
@@ -876,38 +924,38 @@ printsub(direction, pointer, length)
 	  }
 
 	case TELOPT_XDISPLOC:
-	    sprintf(nfrontp, "X-DISPLAY-LOCATION ");
+	    net_printf("X-DISPLAY-LOCATION ");
 	    nfrontp += strlen(nfrontp);
 	    switch (pointer[1]) {
 	    case TELQUAL_IS:
-		sprintf(nfrontp, "IS \"%.*s\"", length-2, (char *)pointer+2);
+		net_printf("IS \"%.*s\"", length-2, (char *)pointer+2);
 		break;
 	    case TELQUAL_SEND:
-		sprintf(nfrontp, "SEND");
+		net_printf("SEND");
 		break;
 	    default:
-		sprintf(nfrontp, "- unknown qualifier %d (0x%x).",
+		net_printf("- unknown qualifier %d (0x%x).",
 				pointer[1], pointer[1]);
 	    }
 	    nfrontp += strlen(nfrontp);
 	    break;
 
 	case TELOPT_NEW_ENVIRON:
-	    sprintf(nfrontp, "NEW-ENVIRON ");
+	    net_printf("NEW-ENVIRON ");
 	    goto env_common1;
 	case TELOPT_OLD_ENVIRON:
-	    sprintf(nfrontp, "OLD-ENVIRON");
+	    net_printf("OLD-ENVIRON");
 	env_common1:
 	    nfrontp += strlen(nfrontp);
 	    switch (pointer[1]) {
 	    case TELQUAL_IS:
-		sprintf(nfrontp, "IS ");
+		net_printf("IS ");
 		goto env_common;
 	    case TELQUAL_SEND:
-		sprintf(nfrontp, "SEND ");
+		net_printf("SEND ");
 		goto env_common;
 	    case TELQUAL_INFO:
-		sprintf(nfrontp, "INFO ");
+		net_printf("INFO ");
 	    env_common:
 		nfrontp += strlen(nfrontp);
 		{
@@ -915,25 +963,25 @@ printsub(direction, pointer, length)
 		    for (i = 2; i < length; i++ ) {
 			switch (pointer[i]) {
 			case NEW_ENV_VAR:
-			    sprintf(nfrontp, "\" VAR " + noquote);
+			    net_printf("\" VAR " + noquote);
 			    nfrontp += strlen(nfrontp);
 			    noquote = 2;
 			    break;
 
 			case NEW_ENV_VALUE:
-			    sprintf(nfrontp, "\" VALUE " + noquote);
+			    net_printf("\" VALUE " + noquote);
 			    nfrontp += strlen(nfrontp);
 			    noquote = 2;
 			    break;
 
 			case ENV_ESC:
-			    sprintf(nfrontp, "\" ESC " + noquote);
+			    net_printf("\" ESC " + noquote);
 			    nfrontp += strlen(nfrontp);
 			    noquote = 2;
 			    break;
 
 			case ENV_USERVAR:
-			    sprintf(nfrontp, "\" USERVAR " + noquote);
+			    net_printf("\" USERVAR " + noquote);
 			    nfrontp += strlen(nfrontp);
 			    noquote = 2;
 			    break;
@@ -947,7 +995,7 @@ printsub(direction, pointer, length)
 				}
 				*nfrontp++ = pointer[i];
 			    } else {
-				sprintf(nfrontp, "\" %03o " + noquote,
+				net_printf("\" %03o " + noquote,
 							pointer[i]);
 				nfrontp += strlen(nfrontp);
 				noquote = 2;
@@ -964,31 +1012,31 @@ printsub(direction, pointer, length)
 
 #if	defined(AUTHENTICATION)
 	case TELOPT_AUTHENTICATION:
-	    sprintf(nfrontp, "AUTHENTICATION");
+	    net_printf("AUTHENTICATION");
 	    nfrontp += strlen(nfrontp);
 
 	    if (length < 2) {
-		sprintf(nfrontp, " (empty suboption??\?)");
+		net_printf(" (empty suboption??\?)");
 		nfrontp += strlen(nfrontp);
 		break;
 	    }
 	    switch (pointer[1]) {
 	    case TELQUAL_REPLY:
 	    case TELQUAL_IS:
-		sprintf(nfrontp, " %s ", (pointer[1] == TELQUAL_IS) ?
+		net_printf(" %s ", (pointer[1] == TELQUAL_IS) ?
 							"IS" : "REPLY");
 		nfrontp += strlen(nfrontp);
 		if (AUTHTYPE_NAME_OK(pointer[2]))
-		    sprintf(nfrontp, "%s ", AUTHTYPE_NAME(pointer[2]));
+		    net_printf("%s ", AUTHTYPE_NAME(pointer[2]));
 		else
-		    sprintf(nfrontp, "%d ", pointer[2]);
+		    net_printf("%d ", pointer[2]);
 		nfrontp += strlen(nfrontp);
 		if (length < 3) {
-		    sprintf(nfrontp, "(partial suboption??\?)");
+		    net_printf("(partial suboption??\?)");
 		    nfrontp += strlen(nfrontp);
 		    break;
 		}
-		sprintf(nfrontp, "%s|%s",
+		net_printf("%s|%s",
 			((pointer[3] & AUTH_WHO_MASK) == AUTH_WHO_CLIENT) ?
 			"CLIENT" : "SERVER",
 			((pointer[3] & AUTH_HOW_MASK) == AUTH_HOW_MUTUAL) ?
@@ -996,26 +1044,26 @@ printsub(direction, pointer, length)
 		nfrontp += strlen(nfrontp);
 
 		auth_printsub(&pointer[1], length - 1, buf, sizeof(buf));
-		sprintf(nfrontp, "%s", buf);
+		net_printf("%s", buf);
 		nfrontp += strlen(nfrontp);
 		break;
 
 	    case TELQUAL_SEND:
 		i = 2;
-		sprintf(nfrontp, " SEND ");
+		net_printf(" SEND ");
 		nfrontp += strlen(nfrontp);
 		while (i < length) {
 		    if (AUTHTYPE_NAME_OK(pointer[i]))
-			sprintf(nfrontp, "%s ", AUTHTYPE_NAME(pointer[i]));
+			net_printf("%s ", AUTHTYPE_NAME(pointer[i]));
 		    else
-			sprintf(nfrontp, "%d ", pointer[i]);
+			net_printf("%d ", pointer[i]);
 		    nfrontp += strlen(nfrontp);
 		    if (++i >= length) {
-			sprintf(nfrontp, "(partial suboption??\?)");
+			net_printf("(partial suboption??\?)");
 			nfrontp += strlen(nfrontp);
 			break;
 		    }
-		    sprintf(nfrontp, "%s|%s ",
+		    net_printf("%s|%s ",
 			((pointer[i] & AUTH_WHO_MASK) == AUTH_WHO_CLIENT) ?
 							"CLIENT" : "SERVER",
 			((pointer[i] & AUTH_HOW_MASK) == AUTH_HOW_MUTUAL) ?
@@ -1027,7 +1075,7 @@ printsub(direction, pointer, length)
 
 	    case TELQUAL_NAME:
 		i = 2;
-		sprintf(nfrontp, " NAME \"");
+		net_printf(" NAME \"");
 		nfrontp += strlen(nfrontp);
 		while (i < length)
 		    *nfrontp += pointer[i++];
@@ -1036,7 +1084,7 @@ printsub(direction, pointer, length)
 
 	    default:
 		    for (i = 2; i < length; i++) {
-			sprintf(nfrontp, " ?%d?", pointer[i]);
+			net_printf(" ?%d?", pointer[i]);
 			nfrontp += strlen(nfrontp);
 		    }
 		    break;
@@ -1046,85 +1094,85 @@ printsub(direction, pointer, length)
 
 #ifdef	ENCRYPTION
 	case TELOPT_ENCRYPT:
-	    sprintf(nfrontp, "ENCRYPT");
+	    net_printf("ENCRYPT");
 	    nfrontp += strlen(nfrontp);
 	    if (length < 2) {
-		sprintf(nfrontp, " (empty suboption??\?)");
+		net_printf(" (empty suboption?)");
 		nfrontp += strlen(nfrontp);
 		break;
 	    }
 	    switch (pointer[1]) {
 	    case ENCRYPT_START:
-		sprintf(nfrontp, " START");
+		net_printf(" START");
 		nfrontp += strlen(nfrontp);
 		break;
 
 	    case ENCRYPT_END:
-		sprintf(nfrontp, " END");
+		net_printf(" END");
 		nfrontp += strlen(nfrontp);
 		break;
 
 	    case ENCRYPT_REQSTART:
-		sprintf(nfrontp, " REQUEST-START");
+		net_printf(" REQUEST-START");
 		nfrontp += strlen(nfrontp);
 		break;
 
 	    case ENCRYPT_REQEND:
-		sprintf(nfrontp, " REQUEST-END");
+		net_printf(" REQUEST-END");
 		nfrontp += strlen(nfrontp);
 		break;
 
 	    case ENCRYPT_IS:
 	    case ENCRYPT_REPLY:
-		sprintf(nfrontp, " %s ", (pointer[1] == ENCRYPT_IS) ?
+		net_printf(" %s ", (pointer[1] == ENCRYPT_IS) ?
 							"IS" : "REPLY");
 		nfrontp += strlen(nfrontp);
 		if (length < 3) {
-		    sprintf(nfrontp, " (partial suboption??\?)");
+		    net_printf(" (partial suboption??\?)");
 		    nfrontp += strlen(nfrontp);
 		    break;
 		}
 		if (ENCTYPE_NAME_OK(pointer[2]))
-		    sprintf(nfrontp, "%s ", ENCTYPE_NAME(pointer[2]));
+		    net_printf("%s ", ENCTYPE_NAME(pointer[2]));
 		else
-		    sprintf(nfrontp, " %d (unknown)", pointer[2]);
+		    net_printf(" %d (unknown)", pointer[2]);
 		nfrontp += strlen(nfrontp);
 
 		encrypt_printsub(&pointer[1], length - 1, buf, sizeof(buf));
-		sprintf(nfrontp, "%s", buf);
+		net_printf("%s", buf);
 		nfrontp += strlen(nfrontp);
 		break;
 
 	    case ENCRYPT_SUPPORT:
 		i = 2;
-		sprintf(nfrontp, " SUPPORT ");
+		net_printf(" SUPPORT ");
 		nfrontp += strlen(nfrontp);
 		while (i < length) {
 		    if (ENCTYPE_NAME_OK(pointer[i]))
-			sprintf(nfrontp, "%s ", ENCTYPE_NAME(pointer[i]));
+			net_printf("%s ", ENCTYPE_NAME(pointer[i]));
 		    else
-			sprintf(nfrontp, "%d ", pointer[i]);
+			net_printf("%d ", pointer[i]);
 		    nfrontp += strlen(nfrontp);
 		    i++;
 		}
 		break;
 
 	    case ENCRYPT_ENC_KEYID:
-		sprintf(nfrontp, " ENC_KEYID %d", pointer[1]);
+		net_printf(" ENC_KEYID %d", pointer[1]);
 		nfrontp += strlen(nfrontp);
 		goto encommon;
 
 	    case ENCRYPT_DEC_KEYID:
-		sprintf(nfrontp, " DEC_KEYID %d", pointer[1]);
+		net_printf(" DEC_KEYID %d", pointer[1]);
 		nfrontp += strlen(nfrontp);
 		goto encommon;
 
 	    default:
-		sprintf(nfrontp, " %d (unknown)", pointer[1]);
+		net_printf(" %d (unknown)", pointer[1]);
 		nfrontp += strlen(nfrontp);
 	    encommon:
 		for (i = 2; i < length; i++) {
-		    sprintf(nfrontp, " %d", pointer[i]);
+		    net_printf(" %d", pointer[i]);
 		    nfrontp += strlen(nfrontp);
 		}
 		break;
@@ -1132,19 +1180,68 @@ printsub(direction, pointer, length)
 	    break;
 #endif	/* ENCRYPTION */
 
+#ifdef TLS
+	case TELOPT_START_TLS:
+	    net_printf("START-TLS ");
+	    nfrontp += strlen(nfrontp);
+	    switch (pointer[1]) {
+	    case TLS_FOLLOWS:
+		net_printf("FOLLOWS");
+		break;
+	    default:
+		net_printf("- unknown qualifier %d (0x%x).",
+				pointer[1], pointer[1]);
+	    }
+	    nfrontp += strlen(nfrontp);
+	    break;
+#endif /* TLS */
+#ifdef FWD_X
+	case TELOPT_FORWARD_X:
+	    net_printf("FORWARD-X ");
+	    nfrontp += strlen(nfrontp);
+	    switch (pointer[1]) {
+	    case FWDX_SCREEN:
+		net_printf("SCREEN %d", pointer[2]);
+		break;
+	    case FWDX_OPEN:
+		net_printf("OPEN %d %d", pointer[2], pointer[3]);
+		break;
+	    case FWDX_CLOSE:
+		net_printf("CLOSE %d %d", pointer[2], pointer[3]);
+		break;
+	    case FWDX_DATA:
+		net_printf("DATA %d %d [%d bytes]", pointer[2], pointer[3], length - 4);
+		break;
+	    case FWDX_OPTIONS: {
+	    	int n;
+		net_printf("OPTIONS");
+		for (n = 2; n < length; n++) {
+		    nfrontp += strlen(nfrontp);
+		    net_printf(" %d", pointer[n]);
+		    }
+		break;
+		}
+	    default:
+		net_printf("- unknown qualifier %d (0x%x).",
+				pointer[1], pointer[1]);
+	    }
+	    nfrontp += strlen(nfrontp);
+	    break;
+#endif /* FWD_X */
+
 	default:
 	    if (TELOPT_OK(pointer[0]))
-		sprintf(nfrontp, "%s (unknown)", TELOPT(pointer[0]));
+		net_printf("%s (unknown)", TELOPT(pointer[0]));
 	    else
-		sprintf(nfrontp, "%d (unknown)", pointer[i]);
+		net_printf("%d (unknown)", pointer[0]);
 	    nfrontp += strlen(nfrontp);
 	    for (i = 1; i < length; i++) {
-		sprintf(nfrontp, " %d", pointer[i]);
+		net_printf(" %d", pointer[i]);
 		nfrontp += strlen(nfrontp);
 	    }
 	    break;
 	}
-	sprintf(nfrontp, "\r\n");
+	net_printf("\r\n");
 	nfrontp += strlen(nfrontp);
 }
 
@@ -1162,15 +1259,15 @@ printdata(tag, ptr, cnt)
 
 	while (cnt) {
 		/* flush net output buffer if no room for new data) */
-		if ((&netobuf[BUFSIZ] - nfrontp) < 80) {
+		if ((&netobuf[netobuf_size] - nfrontp) < 80) {
 			netflush();
 		}
 
 		/* add a line of output */
-		sprintf(nfrontp, "%s: ", tag);
+		net_printf("%s: ", tag);
 		nfrontp += strlen(nfrontp);
 		for (i = 0; i < 20 && cnt; i++) {
-			sprintf(nfrontp, "%02x", *ptr);
+			net_printf("%02x", *ptr);
 			nfrontp += strlen(nfrontp);
 			if (isprint(*ptr)) {
 				xbuf[i] = *ptr;
@@ -1185,8 +1282,93 @@ printdata(tag, ptr, cnt)
 			ptr++;
 		}
 		xbuf[i] = '\0';
-		sprintf(nfrontp, " %s\r\n", xbuf );
+		net_printf(" %s\r\n", xbuf );
 		nfrontp += strlen(nfrontp);
 	}
 }
 #endif /* DIAGNOSTICS */
+
+#define SNPRINTF_OK 1	/* TJW: Should be set by Configure! */
+
+#ifdef __STDC__
+int net_printf(const char *fmt, ...)
+#else
+int net_printf(fmt, va_alist)
+    char *fmt;
+    va_dcl
+#endif
+{
+#define SNP_MAXBUF 1024000
+    va_list ap;
+    register int ret, netobuf_left, nfrontp_idx, nbackp_idx;
+	
+    /* here I boldly assume that snprintf() and vsnprintf() uses the same
+     * return value convention. if not, what kind of libc is this? ;-)
+     */
+#ifdef __STDC__
+    va_start(ap, fmt);
+#else
+    va_start(ap);
+#endif
+    nfrontp_idx = nfrontp - netobuf;	/* equal to how filled netobuf is */
+    nbackp_idx = nbackp - netobuf;
+    netobuf_left = netobuf_size - (nfrontp - netobuf);
+
+    ret = vsnprintf(nfrontp, netobuf_left, fmt, ap);
+#ifdef SNPRINTF_OK
+    /* this one returns the number of bytes it wants to write in case of overflow */
+    if (ret >= netobuf_left) {
+	/* netobuf was too small, increase it */
+	register int new_size = ret + 1 + nfrontp_idx;
+	register char *p;
+	if (new_size > SNP_MAXBUF)
+	    syslog(LOG_ERR, "netobuf_size wanted to go beyond %d", SNP_MAXBUF);
+	else {
+	    p = realloc(netobuf, new_size);
+	    if (p) {
+		netobuf = p;
+		nfrontp = p + nfrontp_idx;
+		nbackp = p + nbackp_idx;
+		netobuf_size = new_size;
+		vsnprintf(nfrontp, ret + 1, fmt, ap);
+	    }
+	}
+    }
+#else
+# ifdef SNPRINTF_HALFBROKEN
+    /* this one returns the number of bytes written (excl. \0) in case of overflow */
+#  define SNP_OVERFLOW(x, y) ( x == y ? 1 : 0 )
+#  define SNP_NOERROR(x)     ( x < 0 ? 0 : 1 )
+# else
+#  ifdef SNPRINTF_BROKEN
+    /* this one returns -1 in case of overflow */
+#   define SNP_OVERFLOW(x, y) ( x < 0 ? 1 : 0 )
+#   define SNP_NOERROR(x)     ( 1 )  /* if -1 means overflow, what's the error indication? */
+#  else
+#   error No valid SNPRINTF_... macro defined!
+#  endif /* !SNPRINTF_BROKEN */
+# endif /* !SNPRINTF_HALFBROKEN */
+    if (SNP_NOERROR(ret) && SNP_OVERFLOW(ret, netobuf_left - 1)) {
+	/* netobuf was too small, increase it */
+	register int new_size = netobuf_size;
+	register char *p;
+	do {
+	    if ((new_size *= 2) > SNP_MAXBUF) {	/* try to double the size */
+		syslog(LOG_ERR, "netobuf_size wanted to go beyond %d", SNP_MAXBUF);
+		break;
+	    }
+	    p = realloc(netobuf, new_size);
+	    if (p) {
+		netobuf = p;
+		nfrontp = p + nfrontp_idx;
+		nbackp = p + nbackp_idx;
+		netobuf_size = new_size;
+		netobuf_left = netobuf_size - (nfrontp - netobuf);
+		ret = vsnprintf(nfrontp, netobuf_left, fmt, ap);
+	    } else
+		break;
+	} while (SNP_NOERROR(ret) && SNP_OVERFLOW(ret, netobuf_left - 1));
+    }
+#endif /* !SNPRINTF_OK */
+    return strlen(nfrontp);
+}
